# coding: utf8
from __future__ import absolute_import
from farbox_lite.utils import UnicodeWithAttrs, smart_unicode, get_md5
from farbox_lite.utils.html import linebreaks, html_to_text, limit as _limit, html_escape
from farbox_lite.utils.lazy import is_str
from farbox_lite.utils.date import date_parse
from farbox_lite.utils.lazy import count_words
from farbox_lite.template_system.functions.attr_patches.models.date import Date



def date(obj): # 转为日期格式
    try:
        return Date(date_parse(obj))
    except:
        return obj


def length(obj):
    return len(obj)


def words(obj):
    if isinstance(obj, (str, unicode)):
        obj = smart_unicode(obj)
        if len(obj) > 100000:
            return -1 # 超过10w字节, 不能处理了
        return count_words(obj)
    else:
        return 0


def escaped(obj):
    # 确保 html 的标签进行转义
    return html_escape(obj)

def plain_html(obj):
    # 先将 html 转为文本，再分行
    return linebreaks(html_to_text(obj))

def to_html(obj):
    # 单纯 html 分行
    return linebreaks(obj)

def to_html_with_images(obj):
    return linebreaks(obj, render_markdown_image=True)

def plain_text(obj):
    # 就是 html 转为 text
    return html_to_text(obj)


def plain(obj):
    # 就是 html 转为 text
    return html_to_text(obj)


def limit(obj, length=None, mark='......', keep_images=True, words=None, remove_a=False, keep_a_html=False, ignore_first_tag_name='blockquote'):
    return _limit(obj, length, mark, keep_images, words, remove_a=remove_a, keep_a_html=keep_a_html, ignore_first_tag_name=ignore_first_tag_name)


def has(obj, to_check, just_in=False):
    if isinstance(to_check, (tuple, list)): # 给的是一个list，只要其中一个元素满足就可以了
        for sub_to_check in to_check[:100]: # 最多检测100项
            if isinstance(sub_to_check, (unicode, str)) and has(obj, sub_to_check, just_in=just_in):
                return True

    if not isinstance(obj, (unicode, str)):
        return False
    if not isinstance(to_check, (unicode, str)):
        return False
    to_check = to_check.strip().lower()
    obj = obj.strip().lower()
    to_check_start = to_check.strip()+' '
    to_check_end = ' ' + to_check.strip()
    if just_in:
        return to_check in obj
    elif is_str(to_check):
        if obj == to_check or obj.startswith(to_check_start) or obj.endswith(to_check_end):
            return True
        else:
            return False
    else: # 中文的，没有空格
        return to_check in obj




def md5(obj):
    return get_md5(obj)


# coding: utf8
from __future__ import absolute_import
from farbox_lite.utils.lazy import to_int, to_float
import re


special_grid_factors = {
    0.2: '1/5',
    0.4: '2/5',
    0.6: '3/5',
    0.8: '4/5',
}
def get_grid_factor(k, base=24, odd_base=5):
    # 一般有两套，比如 1/5 就对应不到 24 中
    k = special_grid_factors.get(k) or k

    if not k:
        return k, base

    if isinstance(k, (str,unicode)):
        if '-' in k or '/' in k:
            v1, v2 = re.split(r'-|/', k, maxsplit=1)
            v1 = to_int(v1)
            v2 = to_int(v2)
            if v1 and v2:
                small_one = min(v1, v2)
                big_one = max(v1, v2)
                if big_one == odd_base:
                    base = odd_base
                k = float(small_one)/float(big_one)
        elif '.' in k:
            k = to_float(k)
        else: # 整数
            k = to_int(k) or 1

    if k and isinstance(k, int) and k>1:
        k = float(k)/base

    if isinstance(k, (float, int)) and k<=1:
        # 之前全部处理为分数了, 这样 1 就是全值了
        k *= base

    # 处理最终的 k 值
    k = to_int(round(k)) or base # 四舍五入并 int
    if k > base:
        k = base
    if k < 1:
        k = 1
    return k, base




def test_get_grid_factor():
    print get_grid_factor(0)
    print get_grid_factor(0.001)
    print get_grid_factor(1)
    print get_grid_factor(0.1)
    print get_grid_factor('0.1')
    print get_grid_factor(0.3)
    print get_grid_factor(0.33)
    print get_grid_factor('1/3')
    print get_grid_factor('1/5')
    print get_grid_factor('1-5')
    print get_grid_factor('2-5')
    print get_grid_factor('5-5')
    print get_grid_factor('2-3')
    print get_grid_factor('2、3') # 会被当做整个单位，因为无法识别
    print get_grid_factor('2/3')


if __name__ == '__main__':
    test_get_grid_factor()


#coding: utf8
from __future__ import absolute_import
import operator
from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.lazy import get_value_from_data, split_list, to_int
from farbox_lite.utils.web.cache_for_function import cache_result



def all_has(objects, key, check_value=None, opt=None):
    """objects是一个list，内个element都是dict，并且都有指定key对应的值"""
    if not objects:
        return False
    for obj in objects[:200]: # 最多200个元素
        if not isinstance(obj, dict):
            return False
        if check_value is None and obj.get(key) is None: # 没有设定check_value，则值为None，则匹配失败
            return False
        elif check_value is not None:
            value_got = obj.get(key)
            if isinstance(value_got, (tuple,list)): # 如果得到的结果是列表，则比对的是其元素长度，因为列表本身没有比较的意义
                value_got = len(value_got)
            if (opt is None or opt=='==') and value_got!=check_value: # 没有设定opt，则只要不等于则匹配失败
                return  False
            elif opt=='>=' and value_got<check_value: # 期望是value_got >= check_value, 结果value_got < check_value，匹配失败
                return False
            elif opt=='<=' and value_got>check_value:
                return False
            elif opt=='>' and value_got<=check_value:
                return False
            elif opt=='<' and value_got>=check_value:
                return False
    return True


def _list_func_wrapper(func):
    def _func(objects, *args, **kwargs):
        if not isinstance(objects, (list, tuple)):
            return objects
        else:
            # 对函数的计算结果进行缓存
            cache_key = '%s:%s:%s' % (id(objects), func.func_name, smart_str(list(args)+kwargs.values()))
            cached_func = cache_result(func, cache_key=cache_key)
            return cached_func(objects, *args, **kwargs)
    return _func




##### sort_by

def sort(objects, attr='position', ordered_attr=None, ordered_keys=None, match=False):
    if attr.startswith('-'):
        reverse = True
        attr = attr.lstrip('-')
    else:
        reverse = False
    if reverse == '-':
        reverse = True
    if attr:
        new_objects = sorted(objects, key=lambda o: get_value_from_data(o, attr))
    else:
        new_objects = objects # 原始的就可以了。
    if reverse:
        new_objects.reverse()

    if ordered_attr and ordered_keys and isinstance(ordered_attr, (str,unicode)) and isinstance(ordered_keys, (list, tuple)):
        ordered_keys = [k for k in ordered_keys if isinstance(k, (str,unicode))]
        objects_map = {get_value_from_data(obj, ordered_attr):obj for obj in new_objects}
        head_objects = []
        for key in ordered_keys:
            key_matched_obj = objects_map.get(key)
            if key_matched_obj is not None:
                head_objects.append(key_matched_obj)
                try: new_objects.remove(key_matched_obj)
                except: pass
        if match: # 仅仅处理 ordered_keys 的逻辑
            return head_objects
        return head_objects + new_objects

    return new_objects




##### group
def _get_sort_key(value, per=10):
    # 将一个value转为可排序的key 值
    # value通常是一个list, per作为阶乘积， 比如[5,2] -> 5*10+2*1= 52
    # 比如年、月、日
    # 主要在group函数中，处理attr:sub1+sub2的这种方式，比如date:year+month
    if type(value) in [tuple, list]:
        is_number = True
        ls = value
        for row in ls:
            if type(row) not in [float, int]:
                is_number = False
                break
        if is_number:
            total = 0
            pos = len(ls)
            for i in ls:
                total += i * per**(pos-1)
                pos -= 1
            return total
        else:
            return ','.join([str(row) for row in ls])
    else:
        return value


def group(objects, attr, *args, **kwargs):
    """
    list
    """
    if attr.startswith('-'):
        reverse = True
        attr = attr.lstrip('-')
    else:
        reverse = kwargs.pop('reverse', False)

    group_sorts = kwargs.pop('group_sorts', None) or []

    key = smart_str(attr)
    parts = key.split(':', 1)
    attr = parts[0]
    sub_attrs = [row.strip() for row in parts[1].split('+')] if len(parts) == 2 else [] # : 表示有子属性，由+进行分割多个字段

    # 使用相同的key值，grouped存objects, grouped_sort存排序的key
    grouped = {}
    grouped_sort = {}

    def add_source(group_name): # 添加源到一个 group 上
        grouped.setdefault(group_name,[]).append(obj)  # 组别, 然后添加到这个 group 中
        if group_name not in grouped_sort:  # 组别的排序, sort_key 是一个可排序的，一般会处理为整数
            grouped_sort[group_name] = sort_key

    for obj in objects:
        if not sub_attrs:  # 没有子属性复合的，直接属性, 比如 group(posts, 'date')， 或者 group(posts, 'date.year')
            sort_key = sort_value = get_value_from_data(obj, attr)  # 不一定是basestring类型的
            if type(sort_value) in [list, tuple]:  # 目前是tags的处理
                # 如果sort_key是一个list，则里面每一个都会分组
                for group_name in sort_value:
                    add_source(group_name)
            else:
                add_source(group_name=sort_value)
        else: # 很可能是复合属性
            parent = get_value_from_data(obj, attr)
            sort_values = []
            for sub_attr in sub_attrs:
                sort_values.append(get_value_from_data(parent, sub_attr))
            sort_values = tuple([value for value in sort_values if value])
            if len(sort_values) == 1:
                sort_values = sort_values[0]
            elif not sort_values:
                sort_values = ''  # 避免返回()这样的值

            sort_key = _get_sort_key(sort_values, per=100)

            add_source(group_name=sort_values) # 以values作为group_name

    # 得到根据sort_key排序了的group_names相关的二元数组
    grouped_sort = sorted(grouped_sort.iteritems(), key=operator.itemgetter(1), reverse=reverse)
    # 获得已经排序成功的group_names
    sorted_group_names = [group_name for group_name, sort_key in grouped_sort]
    # 返回最终结果

    to_return = []
    sorted_groups = []
    unsorted_groups = []
    for group_name in sorted_group_names:
        one_group = [group_name, grouped.get(group_name)]
        to_return.append(one_group)
        if group_sorts:
            if group_name in group_sorts:
                sorted_groups.append(one_group)
            else:
                unsorted_groups.append(one_group)

    if group_sorts:
        sorted_groups.sort(lambda x,y: group_sorts.index(x[0])-group_sorts.index(y[0]))
        to_return = sorted_groups + unsorted_groups

    return to_return






def insert(objects, to_insert, position=-1): # 默认插入最后
    #if isinstance(to_insert, (list, tuple)):
        #return ''
    if isinstance(to_insert, dict) and to_insert.get('_id'):
        to_insert_id = to_insert.get('_id')
        for obj in objects:
            if isinstance(obj, dict) and obj.get('_id') == to_insert_id:
                objects.remove(obj)
                break
    if to_insert:
        try:
            objects.insert(position, to_insert)
        except:
            objects.append(to_insert)
    return '' # 对objects的id没有操作，返回空值



#### filter
def __filter(objects, attr, attr_value, opt=None, return_one=False):
    # posts.filter_by_date('2013-12-14'.date, '>=')
    # posts.filter_by_tag('test', 'in')
    if len(objects) > 2000: # 最大尺寸限制
        objects = objects[:2000]
    if isinstance(attr_value, (list, tuple)): # in for list
        result = filter(lambda obj: get_value_from_data(obj, attr) in attr_value, objects)
    elif attr and isinstance(attr, (str, unicode)):
        filtered_objects = []
        for row in objects:
            value_got = get_value_from_data(row, attr)
            if opt == 'in' and isinstance(attr_value, (str, unicode, int, float)):
                if isinstance(value_got, (list, tuple)) and attr_value in value_got:
                    filtered_objects.append(row)
            elif opt in ['>=', '=>'] and value_got>=attr_value:
                filtered_objects.append(row)
            elif opt in ['<=', '=<'] and value_got<=attr_value:
                filtered_objects.append(row)
            elif not opt:
                if isinstance(value_got, (list, tuple)) and not isinstance(attr_value, (list, tuple)):
                    # like filter(posts, 'tag', 'test')
                    if attr_value in value_got:
                        filtered_objects.append(row)
                else:
                    if type(attr_value) == bool: # boolean 类型的特殊处理
                        if attr_value == bool(value_got):
                            filtered_objects.append(row)
                    elif attr_value == value_got:
                        filtered_objects.append(row)
        result = filtered_objects
    else:
        result = []
    if return_one:
        if result:
            return result[0]
        else:
            return  None
    else:
        return result


def join(objects, link_str=' '):
    link_str = smart_unicode(link_str)
    if all([isinstance(obj, (str, unicode, int, float)) for obj in objects]):
        # 内部元素，必须都是字符串类型的
        objects = [smart_unicode(obj) for obj in objects]
        return link_str.join(objects)
    else:
        return objects


def length(objects):
    return len(objects)



def split(objects, per=3):
    per = to_int(per, min_value=1) or 3
    return split_list(objects, per)


for func in [length]:
    func.is_just_property = True


#coding: utf8
from __future__ import absolute_import
import sys
import os
import hashlib
import re
import uuid


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = (str, bytes)
    unicode = str
else:
    string_types = basestring
    unicode = unicode


def to_bytes(s):
    if not isinstance(s, string_types):
        s = smart_unicode(s)
    if isinstance(s, unicode):
        s = s.encode('utf8')
    else:
        try:
            s = str(s)
        except:
            pass
    return s

ENCODINGS = [
    "utf8",
    "gb18030",
    "big5",
    "latin1",
    "ascii"
]
def to_unicode(s):
    if not isinstance(s, string_types):
        try:
            s = unicode(s)
        except:
            try:
                s = str(s)
            except:
                pass
    if not isinstance(s, unicode):
        for encoding in ENCODINGS:
            try:
                s = unicode(s, encoding)
                return s
            except:
                pass
    return s

smart_unicode = to_unicode
smart_str = to_bytes

def get_md5(content, block_size=0): # 1024*1024 2**20, block_size=1Mb
    content = smart_str(content)
    if not block_size:
        return hashlib.md5(content).hexdigest()
    else:
        md5_obj = hashlib.md5()
        n = len(content)
        for i in range(0, n, block_size):
            md5_obj.update(content[i:i+block_size])
        return md5_obj.hexdigest()

class UnicodeWithAttrs(unicode):
    pass


def is_str(content):
    try:
        str(content)
        return True
    except:
        return False


def unique_list(data):
    # 类似 set 功能，但是保留原来的次序
    new_data = []
    for row in data:
        if row not in new_data:
            new_data.append(row)
    return new_data


def string_to_list(value):
    if not value:
        return []
    if isinstance(value, (str, unicode)) and value:
        if ',' in value:
            ls = value.split(',')
        elif u'，' in value:
            ls = value.split(u'，')
        else:
            ls = value.split(' ')
        ls = [item.strip() for item in ls if item]
        return unique_list(ls)
    elif type(value) in [list, tuple]:
        value = unique_list(value)
        return [smart_unicode(row).strip() for row in value if row]
    else:
        return [smart_str(value)]


def same_slash(filepath):
    filepath = filepath.rstrip('/').replace('\\', '/')
    filepath = smart_unicode(filepath)
    return filepath


def get_relative_path(filepath, root, return_name_if_fail=True):
    # without '/' in head/foot
    filepath = same_slash(filepath)
    root = same_slash(root)
    if filepath and root and filepath.startswith(root+'/'):
        return filepath.replace(root, '').strip('/')
    elif filepath == root:
        return ''
    else:
        if return_name_if_fail:
            return os.path.split(filepath)[-1]
        else:
            return filepath



class LazyDict(dict):
    # 非常有用的一个处理，可以让dict 之类的数据，可以直接调用property 的写法
    def __getitem__(self, item):
        try:
            value = dict.__getitem__(self, item)
        except:
            value = LazyDict()
        return value

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __repr__(self):
        return ''

    def __nonzero__(self):
        return 0


class LazyBool(object):
    # 这样在最终调用的时候，避免有True/False的输出，但是又能作为条件的判断
    def __init__(self, value):
        self.value = value
    def __nonzero__(self):
        if self.value:
            return 1
        else:
            return 0
    def __repr__(self):
        return ''
    def __unicode__(self):
        return ''
    def __str__(self):
        return ''


class ErrorInfo(unicode):
    def __nonzero__(self):
        # 这样判定 if 的时候，始终False
        return 0


def make_content_clean(content):
    content = smart_unicode(content)
    content = content.replace(u'\xa0', u' ')
    content = re.sub('\x07|\x08|\x10|\x11|\x12|\x13|\x14|\x15|\x16|\x17|\x18|\x19|\x1a|\x1b|\x1c|\x1d|\x1e|\x1f', '', content)
    return content



def get_uuid():
    return uuid.uuid1().hex


def get_random_html_dom_id():
    return 'd_%s' % get_uuid()


#coding: utf8
from __future__ import absolute_import
import ujson as json
from farbox_lite.utils import smart_str


def json_dumps(obj, indent=None):
    # 主要是 mongodb 上的一些 obj
    if isinstance(obj, dict) and '_id' in obj:
        obj['_id'] = smart_str(obj['_id'])
    return json.dumps(obj, indent=indent)


def json_loads(raw_content):
    # 主要是 mongodb 上的一些 obj
    return json.loads(raw_content)




def make_tree(docs, kept_fields=None):
    if not docs:
        return docs

    #('title', 'position', '_images_count', 'images_count', 'posts_count', '_posts_count', 'real_path')
    if kept_fields and isinstance(kept_fields, (list, tuple)): # 为避免产生的 tree 包含太多数据，如果指定了 kept_fields
        kept_fields = set(list(kept_fields) + ['slash_number', 'path'])
        _docs = []
        for doc in docs:
            doc = {field:doc.get(field) for field in kept_fields}
            _docs.append(doc)
        docs = _docs

    leveled_docs = {}
    for doc in docs:
        slash_number = doc.get('slash_number') # 肯定是整数
        if slash_number is None:
            continue
        leveled_docs.setdefault(slash_number, []).append(doc)

    levels = leveled_docs.keys()
    if not levels:
        return docs

    min_level = min(levels)
    max_level = max(levels)

    level1_docs = leveled_docs.pop(min_level) # 先得到第一级的
    if min_level == max_level: # 没有多层结构
        return level1_docs

    parent_level_docs = level1_docs
    for level in range(min_level+1, max_level+1): # 得到后续的 level
        current_level_docs = leveled_docs.pop(level, [])
        if not current_level_docs:
            break
        parent_match_dict = {doc['path']:doc for doc in parent_level_docs}
        for doc in current_level_docs:
            parent_path = doc['path'].rsplit('/', 1)[0]
            parent_doc = parent_match_dict.get(parent_path)
            if parent_doc and isinstance(parent_doc, dict):
                children = parent_doc.setdefault('children', [])
                if doc not in children:
                    children.append(doc)
                #parent_doc.setdefault('children', []).append(doc)
        parent_level_docs = current_level_docs
    return level1_docs



#coding: utf8
from __future__ import absolute_import
from collections import OrderedDict
from farbox_lite.utils import get_md5, smart_str, get_relative_path
from farbox_lite.utils.functional import curry
import copy

class LimitedSizeDict(OrderedDict):
    def __init__(self, *args, **kwds):
        self.size_limit = None # default，实际上不会有 limited 的效果
        fields = ['size_limit', 'limit', 'size', 'max', 'max_size']
        for field in fields:
            field_value = kwds.pop(field, None) # 避免这些字段成为字段的初始对象
            if field_value:
                self.size_limit = field_value
        OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key, value, *args, **kwargs):
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self):
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=True)


class cached(object):
    """
    @cached(get_cache_key_func)
    def xxx(xxxxx):
        xxxxxxx
    """
    def __init__(self, cache_key_func):
        self.cached_function_responses = LimitedSizeDict(size=1000)
        self.cache_key_func = cache_key_func


    def __call__(self, func):
        prefix_cache_key = self.cache_key_func()
        def _func(*func_args, **func_kwargs):
            base_cache_key = '%s-%s' % (prefix_cache_key, id(func))
            if func_args or func_kwargs:
                # 可能第一个变量传入的是 self 这种实例对象
                func_args_s_list = [] # 形成缓存key用的一个list
                for func_arg in func_args:
                    if not isinstance(func_arg, (str, unicode, dict, tuple, list, bool, int, float)):
                        func_args_s_list.append(smart_str(id(func_arg)))
                    else:
                        if isinstance(func_arg, dict) and func_arg.get('id'):
                            func_args_s_list.append(func_arg.get('id'))
                        else:
                            func_args_s_list.append(func_arg)
                if func_kwargs:
                    func_kwargs = copy.copy(func_kwargs)
                    for k, v in func_kwargs.items():
                        if v and isinstance(v, dict) and v.get('id'):
                            func_kwargs[k] = v.get('id')
                var_key = '-'.join(func_args_s_list) + smart_str(func_kwargs)
                var_md5 = get_md5(var_key)
                cache_key = '%s-%s' % (base_cache_key, var_md5)
            else:
                cache_key = base_cache_key

            if cache_key in self.cached_function_responses:
                cached_data = self.cached_function_responses[cache_key]
                return cached_data
            else:

                result = func(*func_args, **func_kwargs)
                self.cached_function_responses[cache_key] = result
                return result
        return _func



def get_sub_docs_by_path(root, raw_docs, parent_path, is_direct=False):
    if parent_path.startswith('/'):
        parent_relative_path = get_relative_path(parent_path, root=root).lower()
    else:
        parent_relative_path = parent_path.strip().strip('/').lower()
    parent_slash_number = parent_relative_path.count('/')
    docs = []
    for doc in raw_docs:
        doc_path = doc.get('path')
        if not doc_path:
            continue
        doc_slash_number = doc_path.count('/')
        if not doc_path.startswith('%s/'%parent_relative_path):
            continue
        if is_direct:
            if doc_slash_number != parent_slash_number+1:
                continue
        docs.append(doc)
    return docs


def get_sub_docs_by_field(raw_docs, field, value):
    docs = []
    for doc in raw_docs:
        if not isinstance(doc, dict):
            continue
        if doc.get(field) == value:
            docs.append(doc)
    return docs



def sort_key_func(field, doc):
    return doc.get(field)


def sort_docs_by_field(raw_docs, field=None):
    if not field:
        return raw_docs
    field = field.strip()
    if field.startswith('-'):
        reverse = True
    else:
        reverse = False
    field = field.strip('-').strip()
    sorted_docs = sorted(raw_docs, key=curry(sort_key_func, field), reverse=reverse)
    return sorted_docs




# coding: utf8
from __future__ import absolute_import



# 不支持自适应的逻辑

def re_key_for_layout_element(element, base=24):
    # 24 作为grid的基准，然后 0.33  1/4 等转为24为分母的值
    if not isinstance(element, dict) or not element.get('key'):
        return
    key = element.get('key')
    new_key = None
    if isinstance(key, (float, int)):
        if 0<key<=1: # 小数点
            new_key = int(round(base*key)) # 四舍五入的整数化
        elif key > base:
            new_key = base
    if new_key:
        element['key'] = new_key
    return ''



def re_key_for_layout_elements(elements, base=24):
    if isinstance(elements, (list, tuple)):
        for element in elements:
            re_key_for_layout_element(element, base)
    return ''



def split_elements_to_rows(elements, base=24):
    # 计算用于布局的 elements，这样最终在模板中渲染的时候，就直接调用属性就可以了
    rows = [[]]
    current_width = 0
    if not isinstance(elements, (list, tuple)):
        return rows

    re_key_for_layout_elements(elements, base=base)

    for element in elements:
        if not isinstance(element, dict):
            continue
        layout_key = element.get('key')

        if layout_key in ['h']: # 行 & 高度分割
            current_width = 0
            rows[-1].append(element)
            rows.append([])
            continue

        if isinstance(layout_key, (int, float)):
            new_current_width = current_width + int(layout_key)
            if new_current_width >= base: # 换行了
                current_width = 0
                rows[-1].append(element)
                rows.append([])
                continue
            else: # 还没有换行，当前总宽度增加
                current_width += int(layout_key)

        rows[-1].append(element)


    formatted_rows = [] # 有些元素属性，merge 到 row 上， 比如 bg
    property_fields = ['bg', 'style']
    for row in rows:
        if not row:
            continue
        formatted_row = dict()
        row_elements = []
        for element in row:
            key = element.get('key')
            if key in property_fields:
                formatted_row[key] = element
            else:
                if not key: # element 内的分行
                    if row_elements:
                        last_element = row_elements[-1]
                        last_element.setdefault('line_elements', []).append(element)
                else:
                    row_elements.append(element)

        formatted_row['elements'] = row_elements
        formatted_rows.append(formatted_row)


    return formatted_rows






# coding: utf8
from __future__ import absolute_import
import datetime
from farbox_lite.utils.web.data import json_dumps, json_loads
from farbox_lite.utils.mime import guess_type
from farbox_lite.utils.lazy import to_int, to_float, string_to_int
from farbox_lite.template_system.functions.attr_patches.models.date import Date



def debug_obj(obj):
    # 主要是开发时候debug用的，可以确定obj是什么类型的对象
    return obj


def auto_value(obj):
    if isinstance(obj, datetime.datetime):
            return Date(obj)
        # at last
    return obj



def json(obj):
    """
    unicode
    {"is_property": true}
    """
    if isinstance(obj, (tuple, list, dict, str, unicode)):
        try:
            return json_dumps(obj, indent=4)
        except:
            return obj
        # todo 缓存策略
        # return cache_result(cache_key)(json_dumps)(obj)
    else:
        return obj

def from_json(obj):
    # 将一个字符串，重新载入为一个dict/list等类型的py 数据
    if not obj:
        return obj
    if not isinstance(obj, (str, unicode)) and hasattr(obj, 'core'):
        # 可能是 Text/ Date 等wrap后的数据对象
        obj = obj.core
    if isinstance(obj, (str, unicode)):
        try:
            return json_loads(obj)
        except:
            return obj
    return obj


def __int(obj):
    if isinstance(obj, (int, float, str, unicode)):
        return to_int(obj, default_if_fail=int)
    return obj

def force_int(obj):
    # 主要是字符串类型，强制提取一个整数出来
    if isinstance(obj, (float, int)):
        return int(obj)
    else:
        return string_to_int(obj)

def __float(obj):
    if isinstance(obj, (int, float, str, unicode)):
        return to_float(obj, default_if_fail=int)
    return obj


def is_int(obj):
    return isinstance(obj, int)

def is_number(obj):
    return isinstance(obj, (int, float))

def is_float(obj):
    return isinstance(obj, float)

def is_str(obj):
    return isinstance(obj, (str, unicode))

def is_unicode(obj):
    return isinstance(obj, (str, unicode))


def is_array(obj):
    return isinstance(obj, (list, tuple))

def is_list(obj):
    return isinstance(obj, list)

def is_tuple(obj):
    return isinstance(obj, tuple)



def is_dict(obj):
    return isinstance(obj, dict)


def is_image(obj): # 一个字符串是否是图片的 url
    if isinstance(obj, (str, unicode)):
        mime_type = guess_type(obj)
        if mime_type and mime_type.startswith('image/'):
            return False
    return False


def type(obj):
    types = {int: 'int', float:'float', bool:'bool', unicode:'str', str:'str', list:'list', tuple:'tuple'}
    for t, ts in types.items():
        if isinstance(obj, t):
            return ts
    # at last
    return ''



def get(obj, *args, **kwargs):
    if isinstance(obj, dict):
        return obj.get(*args, **kwargs)
    elif isinstance(obj, (list, tuple)) and len(args)==1 and isinstance(args[0], int):
        # list 获得第几个元素
        i = args[0]
        try:
            return obj[i]
        except:
            return None
    elif args:
        field = args[0]
        default = args[1] if len(args)>=2 else None
        try:
            return getattr(obj, field, default)
        except:
            return default
    else:
        return # ignore


def absolute(obj):
    if isinstance(obj, (float, int)):
        return abs(obj)
    return obj








#coding: utf8
from __future__ import absolute_import
from farbox_lite.template_system.env import farbox_lite_env
from farbox_lite.render.utils import api_templates



def render_template_source(template_source, *args, **kwargs):
    if not template_source:
        return ''
    try:
        template = farbox_lite_env.from_string(template_source)
        html_content = template.render(*args, **kwargs)
    except:
        html_content = '<div style="color:red">api template error</div>'
    return html_content



def render_api_template(name, *args, **kwargs):
    template_source = getattr(api_templates, name, '')
    return render_template_source(template_source, *args, **kwargs)



#coding: utf8
from __future__ import absolute_import
import os
import uuid

from farbox_lite.compiler.files import get_compiled_file_docs
from farbox_lite.utils import same_slash, smart_unicode, get_relative_path
from .utils import cached, get_sub_docs_by_path, get_sub_docs_by_field, sort_docs_by_field

file_docs_cache = {}

cache_uuid = uuid.uuid4().hex
def get_cache_uuid():
    return cache_uuid

cache_wrapper = cached(cache_key_func=get_cache_uuid)


def get_file_docs(root):
    global cache_uuid
    root = same_slash(root)
    if not os.path.isdir(root):
        return []
    if root in file_docs_cache:
        docs = file_docs_cache.get(root)
        return docs
    docs = get_compiled_file_docs(root) or []
    file_docs_cache[root] = docs
    cache_uuid = uuid.uuid4().hex
    return docs


def update_file_docs(root):
    global cache_uuid
    root = same_slash(root)
    if not os.path.isdir(root):
        return []
    docs = get_compiled_file_docs(root) or []
    file_docs_cache[root] = docs
    cache_uuid = uuid.uuid4().hex
    return docs


@cache_wrapper
def get_file_docs_by_field(root, field, value, sort=None):
    raw_docs = get_file_docs(root)
    docs = get_sub_docs_by_field(raw_docs, field, value)
    docs = sort_docs_by_field(docs, field=sort)
    return docs


@cache_wrapper
def get_file_docs_by_path(root, parent_path, is_direct=False, sort=None):
    raw_file_docs = get_file_docs(root)
    file_docs = get_sub_docs_by_path(root, raw_file_docs, parent_path, is_direct=is_direct)
    file_docs = sort_docs_by_field(file_docs, field=sort)
    return file_docs


@cache_wrapper
def get_image_docs(root, sort=None):
    docs = get_file_docs_by_field(root, 'type', 'image')
    docs = sort_docs_by_field(docs, field=sort)
    return docs


@cache_wrapper
def get_image_docs_by_path(root, parent_path, is_direct=False, sort=None):
    raw_file_docs = get_file_docs(root)
    file_docs = get_image_docs(root, raw_file_docs, parent_path, is_direct=is_direct)
    file_docs = sort_docs_by_field(file_docs, field=sort)
    return file_docs



#coding: utf8
from __future__ import absolute_import
from flask import request, abort, g
from farbox_lite.utils.web.response import force_redirect
from farbox_lite.render.apis.site import get_site_from_request
from farbox_lite.utils.lazy import to_int


def set_basic_vars():
    g.site = get_site_from_request()

    # patch request
    is_ssl = False
    if request.environ.get('HTTP_X_PROTOCOL') == 'https':
        is_ssl = True
        request.url = request.url.replace('http://', 'https://')
        request.url_root = request.url_root.replace('http://', 'https://')
    if request.method in ['HEAD']:
        abort(405)

    g.page = None # 分页的数int型
    if request.view_args and 'page' in request.view_args:
        # 当前第几页，这个是一个全局的机制
        g.page = to_int(request.view_args.get('page'), 0)


#coding: utf8
# 没有放在website.utils中，是因为这两个函数，在backend中会被用到
import base64
from flask import request, escape, g
import re, urlparse
import user_agents
from farbox_lite.utils import smart_unicode





def cache_in_request(func):
    # 缓存到 g 这个空间中，但是要注意变量名的前缀，不要冲突
    def _func(*args, **kwargs):
        attr_name = func.__name__
        if attr_name.startswith('get_'):
            attr_name = attr_name.replace('get_', '', 1)
        if not hasattr(g, attr_name):
            attr_value = func()
            setattr(g, attr_name, attr_value)
        return getattr(g, attr_name, None)
    return _func




def has_file_in_request():
    """request中是否有文件内容"""
    if request.files or 'raw_content' in request.form or 'base64' in request.form:
        return True
    elif 'file' in request.form:
        return True
    else:
        return False


def get_content_in_file():
    """提取request的文件内容，常和has_file_in_request配合使用"""
    if hasattr(g, 'cached_file_content'): # 已经处理过了，不重复处理
        return g.cached_file_content
    if 'file' in request.form:
        # txt会被requests发送的时候，当做POST内的数据处理
        # 也包括中文名的PDF文件.. 因为是一起发送出来的，主要是客户端的上传、同步逻辑的对应
        file_content = request.form['file']
    elif 'file' in request.files:
        file_content = request.files['file'].read()
    elif request.files: # 只取一个文件
        file_content = request.files[request.files.keys()[0]].read()
    elif 'raw_content' in request.form:
        file_content = request.form.get('raw_content')
    elif 'base64' in request.form:
        content = request.form.get('base64')
        if ',' in content[:50]:
            head, content = content.split(',', 1)
        try:
            file_content = base64.b64decode(content)
        except:
            file_content = ''
    else:
        file_content = ''
    g.cached_file_content = file_content
    return file_content


def has_args(args):
    for arg in args:
        if not request.args.get(arg):
            return False
    return True

def get_args(args, lower=False):
    values = []
    for arg in args:
        value = request.args.get(arg, '')
        if lower:
            value = value.lower()
        values.append(value)
    return values


def get_language(environ=None):
    # todo 可以在 cookie 中尝试也读取, lang
    try:
        if request.args.get('lang'): # 可以通过GET的防止强制指定
            lang = request.args.get('lang')
            return lang.replace('-', '_')
        environ = environ or request.environ
    except RuntimeError:
        return ''
    raw_lang = environ.get('HTTP_ACCEPT_LANGUAGE', '')
    if not raw_lang:
        return ''
    lang_c = re.search(r'[a-z]{2}-[a-z]{2}', raw_lang , re.I)
    if lang_c:
        lang = lang_c.group().lower()
    else:
        lang = ''
    return lang.replace('-', '_')



def get_referrer_host(referrer=None):
    try:
        referrer = referrer or request.referrer
    except RuntimeError:
        return ''
    if referrer:
        try:
            host = urlparse.urlparse(referrer).netloc
            return host.lower()
        except:
            return ''
    else:
        return ''


def get_user_agent():
    if hasattr(g, 'user_agent'):
        return g.user_agent
    u = user_agents.parse(request.environ.get('HTTP_USER_AGENT', ''))
    g.user_agent = u
    return u


def safe_get(key): # from request.form
    value = request.form.get(key)
    if value:
        value = smart_unicode(escape(value))
    else:
        value = ''
    if key == 'site':
        if value and '://' not in value:
            value = 'http://' + value
    return value


def set_g_default(key, default):
    if not hasattr(g, 'g_defaults'):
        g.g_defaults = {}
    return g.g_defaults.setdefault(key, default)



def get_visitor_ip():
    # 得到访客的 ip
    if request.remote_addr: # 比如 nginx 过来的，proxy-pass & 走 unix socket 的缘故，remote_addr 会是空的, 而是打到 X-Forwarded-For 上
        return request.remote_addr
    else:
        ip = request.environ.get('HTTP_X_FORWARDED_FOR') or ''
        if ip:
            return ip
        elif request.access_route:
            return request.access_route[-1].strip()
        else:
            return ''


def is_normal_request():
    user_agent = request.headers.get('User-Agent', '')
    if not user_agent:
        return False
    if re.search(r'(python|rails|urllib|ezooms|fetcher|monitor|spider|feed|bot|http|@)', user_agent, re.I) or \
        'MSIE 6.0' in user_agent or re.search(r'referr?er\.', request.host, re.I):
        # 类似爬虫之类的
        return False
    elif  len(user_agent) < 20:
        return False
    # at last
    return True



# coding: utf8
from __future__ import absolute_import
import re
from flask import g
import HTMLParser
from farbox_markdown.meta import extract_metadata

from farbox_lite.utils import smart_unicode
from farbox_lite.utils.functional import cached_property
from farbox_lite.utils.lazy import get_value_from_data, is_on, to_int, to_float, re_sub, cut_content_by_words
from farbox_lite.utils.html import linebreaks, html_to_text, limit as _limit, html_escape


unescape = HTMLParser.HTMLParser().unescape

HTML_C = re.compile(r'</?[^<]+?>')


class Text(object):
    def __init__(self, core, attr=None, parent=None):
        self.parent = parent # text 字段所在的父data，比如post
        self.attr = attr # 该字段被调用的名称
        if not isinstance(core, unicode):
            core = smart_unicode(core)
        self.core = core

        # 对 post 类型做的特别处理
        if self.attr == 'content' and get_value_from_data(self.parent, 'type')=='post':
            self.is_post_content = True
        else:
            self.is_post_content = False


    def __nonzero__(self):
        # 判断 boolean 用的
        if self.core:
            return True
        else:
            return False

    def __eq__(self, other):
        return other == self.core


    def __contains__(self, item):
        if not isinstance(item, (str, unicode)):
            item = smart_unicode(item)
        return item in self.core


    def _get_slice_content(self):
        if self.parent and self.attr=='content' and 'raw_content' in self.parent:
            content_to_slice = self.plain_text
        else:
            content_to_slice = self.core
        return content_to_slice


    def __getitem__(self, length):
        # 截取部分内容的
        if isinstance(length, int):
            content_to_slice = self._get_slice_content()
            return linebreaks(content_to_slice[:length], self.post_path)
        return self.core

    def __call__(self, length): # 同上， 截取部分内容的
        return self.__getitem__(length)


    def __getslice__(self, i, j):
        # 也是截取内容，但是可以指定开始、结束的位置
        content_to_slice = self._get_slice_content()
        return content_to_slice[i: j]


    def __repr__(self):
        return self._auto_content

    def __unicode__(self):
        return self._auto_content

    def __add__(self, other):
        # 对相加的特殊处理
        if other is None:
            other = ''
        return '%s%s' % (self.core, other)


    def count(self, key):
        return self.core.count(key)


    @cached_property
    def int(self):
        # '123'.int  or '123'.int('98')
        return to_int(self.core)


    @cached_property
    def float(self):
        # 类似 int 的用法
        return to_float(self.core)


    @cached_property
    def escaped(self):
        return html_escape(self.core)

    @cached_property
    def plain(self):
        return self.plain_text

    @cached_property
    def clean_text(self):
        # 平文本, 即使原来是 html 类型的，也会转为普通文本; 如果有raw_content, 则取raw_content(去除了metadata声明的部分)
        if self.parent and isinstance(self.parent, dict) and 'raw_content' in self.parent:
            raw_content = self.parent['raw_content'] or ''
            clean_content, metadata = extract_metadata(raw_content) # 去除头部可能的 meta 声明的敏感信息
            return clean_content
        else:
            return html_to_text(self.core)

    @cached_property
    def plain_text(self):
        # 平文本, 即使原来是 html 类型的，也会转为普通文本
        return html_to_text(self.core)


    @cached_property
    def plain_html(self):
        # 仅仅处理图片 & 换行的 html
        return  linebreaks(self.plain_text, post_path=self.post_path, render_markdown_image=True)


    @cached_property
    def _toc_content(self):
        # 主要是 post 的TOC内容
        if not self.parent or not isinstance(self.parent, dict):
            return ''
        toc_content = self.parent.get('toc', '')
        if toc_content:
            return "<div class='toc'>%s</div>" % unescape(toc_content)
        else:
            return ''

    @cached_property
    def _show_toc(self):
        # 只有post，并且post以markdown格式显示的，才有__show_toc这个属性
        if self.is_post_content and self.parent and get_value_from_data(g, 'site.configs.post_content_type') != 'plain':
            # post未设定，走site.configs # post已设定，走post.metadata
            show_toc = get_value_from_data(self.parent, 'metadata.toc', get_value_from_data(g, 'site.configs.show_toc')) # 默认是post自行定义
            return is_on(show_toc)
        # 默认返回None，表示否
        return False


    @cached_property
    def _content(self):
        if self.attr in ['content',] and self.is_post_content and self.parent and self.parent.get('raw_content') and \
                        get_value_from_data(g, 'site.configs.post_content_type')=='plain':
            # post 作为 普通的文本作为解析
            core_content = linebreaks(self.clean_text, post_path=self.post_path, render_markdown_image=True)
            # 增加一个wrap，做一个 class 的外部表示
            core_content = '<div class="plain_content">\n%s\n</div>' % core_content
        else:
            core_content = self.core
        return core_content


    @cached_property
    def _auto_content(self):
        # 主要是直接调用时候的处理，比如 site.title  post.content(这个可能有TOC 或者什么的)
        if self._show_toc and self._toc_content:
            return self._toc_content + self._content
        else:
            return self._content


    @cached_property
    def opening(self):
        # 获取摘要
        opening_parts = re.split(r'<!-- *more *-->', self.core, 1, flags=re.I)
        if len(opening_parts) == 2:
            return opening_parts[0]
        else:
            # 没有分割，就返回空值
            return ''

    @cached_property
    def post_path(self):
        # 日志的路径
        if self.parent and self.parent.get('path') and self.is_post_content:
            return self.parent.get('path')

    def limit(self, length=None, mark='......', keep_images=True, words=None, remove_a=False, keep_a_html=False, ignore_first_tag_name='blockquote'):
        result = _limit(content=self.core, length=length, mark=mark,
                        keep_images=keep_images, words=words, post_path=self.post_path,
                        remove_a=remove_a, keep_a_html=keep_a_html, ignore_first_tag_name=ignore_first_tag_name)
        return result

    # 基本函数的调用 ends



#coding: utf8
import json, urllib
from farbox_lite.utils import smart_str, smart_unicode
from urlparse import urlparse, parse_qs

def get_base_url(url):
    url = url.strip()
    u = urlparse(url)
    return '%s://%s/' % (u.scheme, u.netloc)

def get_url_path(url): # like /xxx
    u = urlparse(url)
    return u.path


def get_host_from_url(url):
    if not url:
        return
    url = url.strip()
    u = urlparse(url)
    result = u.netloc
    if result:
        result = result.lower().strip()
    return result


def is_same_url(url1, url2, check_GET=False):
    if not isinstance(url1, (str, unicode)):
        return False
    if not isinstance(url2, (str, unicode)):
        return False
    if url1.startswith('https://'):
        url1 = url1.replace('https://', 'http://', 1)
    if url2.startswith('https://'):
        url2 = url1.replace('https://', 'http://', 1)
    url1 = smart_unicode(url1)
    url2 = smart_unicode(url2)
    if check_GET:
        return url1==url2
    else: # 构建两个url去除GET参数后的
        url1 = url1.split('?')[0].rstrip('/').rstrip()
        url2 = url2.split('?')[0].rstrip('/').rstrip()
        return url1==url2



def get_get_var(url, key):
    u = urlparse(url)
    query = u.query
    if not query:
        return
    else:
        qs = parse_qs(query)
        value = qs.get(key)
        if isinstance(value, (list, tuple)) and len(value)==1:
            return value[0]
        else:
            return value

def get_GET_dict_data(query_string):
    qs = parse_qs(query_string)
    dict_data = {}
    for k, v in qs.items():
        if isinstance(v, (list, tuple)) and len(v) == 1:
            v = v[0]
        dict_data[k] = v
    return dict_data



def join_url(url, **params):
    if not params:
        return url
    for key in params:
        value = params[key]
        if isinstance(value, dict):
            params[key] = json.dumps(value)
        elif isinstance(value, unicode):
            params[key] = value.encode('utf8')
        elif value is True:
            params[key] = 'true'
        else:
            params[key] = smart_str(value)

    url_parts = url.split('?', 1)
    if len(url_parts) == 2:
        # 合并原来的 GET 参数
        url, url_q = url_parts
        GET_dict_data = get_GET_dict_data(url_q)
        for k, v in GET_dict_data.items():
            if k not in params:
                params[k] = v

    p_strings = urllib.urlencode(params)

    url += '?'+p_strings
    return url


def encode_url_arg(arg):
    if '%' in arg: # 认为已经encode过了的
        return arg
    try:
        return urllib.quote(arg)
    except:
        return arg

def decode_url_arg(arg):
    try:
        return urllib.unquote(arg)
    except:
        return arg


#coding: utf8
from __future__ import absolute_import
import re
from farbox_lite.utils import smart_unicode, UnicodeWithAttrs, make_content_clean
from farbox_lite.utils.functional import curry
from farbox_lite.utils.lazy import get_value_from_data, cut_content_by_words
from farbox_markdown.compile_md import fix_relative_image_path
import cgi

# backend也有用到,所以放在farbox.utils中

def html_escape(content):
    escaped_content = cgi.escape(content)
    escaped_content = make_content_clean(escaped_content)
    # escaped_content = escaped_content.replace('\x08', '').replace('\x07', '') # 去除 \a \b 标记
    return escaped_content

def linebreaks(value, post_path=None, render_markdown_image=False):
    """Converts newlines into <p> and <br />s."""
    # post_path 是处理图片本身的相对路径，这样就可以处理Markdown语法的图片了
    value = re.sub('<.*?[^>]$','', value) #避免尾行图片（源码）被截断
    value = re.sub(r'\r\n|\r|\n', '\n', value)
    paras = re.split('\n{2,}', value)
    new_paras = []
    for p in paras:
        p = "\n\n".join([u"<p>%s</p>"%line for line in re.split('\n',p)])
        p = u'<div class="p_part">%s</div>' %p
        new_paras.append(p)
    html = u'\n\n'.join(new_paras)
    if post_path or render_markdown_image:
        html = re.sub(r'!\[(.*?)\]\(([^")]+)\)', '<img title="\g<1>" src="\g<2>"/>', html) # 将Markdown  图片语法的转为 html
    if post_path:
        html = fix_relative_image_path(post_path, html) # 相对路径的图片需要进行转化
    return html




def html_to_text(content, keep_a_html=False, remove_a=True, keep_images=False, quote_tags=False):
    # keep_a_html 可以保留 a 元素 的完整逻辑
    # remove_a 表示整个A 元素删除，反之则是转为普通的文本
    content = re.sub(r'<!--.*?-->', '', content) # 先去除HTML的注释
    content = re.sub('<div class="linenodiv">.*?</div>', '', content, flags=re.S) # 去代码高亮
    content = re.sub(r'<br\s*/?>', '\n', content, flags=re.I|re.S)
    content = re.sub(r'</p>\s*<p>', '\n\n', content, flags=re.I|re.S)
    if keep_images:
        content = re.sub('(<img)([^<]+)(/?\s*>)', '&LT;img\g<2>/&GT;', content) # 保护图片代码

    if keep_a_html:
        remove_a = False
    if remove_a: # 会把 a 元素整体删除..
        content = re.sub(r'<a[^<]+>([^<]+)</a>', '', content, re.I|re.S)
    if keep_a_html: # 保留原始的 A 元素
        content = re.sub(r'(<)(a[^<]+)(>)([^<]+)(</a>)', '&LT;\g<2>&GT;\g<4>&LT;/a&GT;', content, re.I|re.S)

    content = re.sub(r'</?[^<]+?>', '', content).strip() # 去tag 标签
    #content = html_unescape(content)

    if quote_tags:
        content = content.replace('<', '&lt;').replace('>', '&gt;')


    # 还原 A  & IMG 元素
    content = content.replace('&LT;', '<').replace('&GT;', '>')


    content = re.sub(r'\n +', '\n', content)

    return content


HTML_C = re.compile(r'</?[^<]+?>')
def limit(content, length=None, mark='......', keep_images=True, words=None, post_path=None, remove_a=False, keep_a_html=False, ignore_first_tag_name=None):
    # like patch...
    # smartpage 中声明 css的
    # post_path 指定的，则会尝试解析 Markdown （插图）语法的逻辑
    content = re.sub('<div class="codehilite code_lang_css  highlight">.*?<!--block_code_end-->', '', content, flags=re.S)

    if ignore_first_tag_name:
        # 忽略掉首个 dom 元素，一般是 blockquote
        if isinstance(ignore_first_tag_name, (str, unicode)) and re.match(r'[a-z]+$', ignore_first_tag_name, re.I):
            content = content.strip()
            content = re.sub(r'^<%s[^<>]*>.*?</%s>'%(ignore_first_tag_name, ignore_first_tag_name), '', content, flags=re.I|re.S)

    if HTML_C.search(content): #如果有html的代码，则进行平文本转义
        content = html_to_text(content, keep_images=keep_images, remove_a=remove_a, keep_a_html=keep_a_html, quote_tags=True)
    if not length and not words: # 长度都没有说声明
        output = content
        has_more = False
    else:
        if words and isinstance(words, int):
            output = cut_content_by_words(content, words, mark=mark)
            has_more = getattr(output, 'has_more', False)
        else:
            output = content[0:length]
            if len(content) > length:
                has_more = True
                if mark:
                    output += smart_unicode(mark)
            else:
                has_more = False
    output = linebreaks(output, post_path=post_path)
    content = UnicodeWithAttrs(output)
    content.has_more = has_more
    content.without_pics = re.sub('<img[^<]+/\s*>', '', output) #去掉img节点
    content.without_pic = content.no_pic = content.without_pics

    return content





def replace_content_with_vars_by_re(value, re_obj):
    #all_matched = re_obj.group(0)
    key = re_obj.group(1)
    attrs = re_obj.group(2)
    if attrs: # 是属性的调用
        attrs = attrs.strip().strip('.')
        value = get_value_from_data(value, attrs)
    if value is None: value = ''
    value = smart_unicode(value)
    return value

def update_content_with_vars(content, data):
    # {{ var }}  的替换
    if not isinstance(data, dict):
        return content
    content = smart_unicode(content)
    content = content.replace('%7B%7B', '{{').replace('%7D%7D', '}}') # 转义
    for key, value in data.items():
        if re.match(r'[a-z0-9-_]+$', key): # key like comment, without dot, attr_call like comment.author
            content = re.sub(r'\{\{\s*(%s)([\w.]+)?\s*\}\}'% key, curry(replace_content_with_vars_by_re, value), content, flags=re.I)
    content = re.sub(r'\{\{\s*[\w.]+\s*\}\}', '', content) # 去除潜在未替换的变量
    return content



def get_images_from_html(content, relative_to_site=False):
    raw_images = re.findall("""<\s*img.*?src=['"]([^'"]+).*?>""", content, flags=re.I)
    images = []
    for image_src in raw_images:
        if image_src in images:
            continue
        if relative_to_site and ('://' in image_src or image_src.startswith('//')):
            continue
        images.append(image_src)
    return images


def _re_html_dom(html, rule):
    # 尝试对某些类型的 dom 元素的类型进行修改或者 wrap
    # h2: <section class=t1><section class=t2><section class=t3>$</section></section></section>\n
    # $ 表示作为 innerHTML， $$ 表示作为 outerHTML
    # 比如此列，$ 表示原来 h2 的innerHTML， 而 $$ 则是包括原始的整个h2
    if not isinstance(html, (str, unicode)):
        return html
    if not isinstance(rule, (list, tuple)) or len(rule)!=2:
        return html
    dom_type, dom_template = rule
    if not isinstance(dom_type, (str, unicode)) and not isinstance(dom_template, (str, unicode)): # ignore
        return html
    dom_type = dom_type.strip()
    if not re.match(r'[a-z0-9]{1,20}$', dom_type, flags=re.I): # 不符合dom type的类型，主要是避免产生正则的性能消耗
        return html
    matched_doms = re.findall(r'(<%s[^<>]*>(.*?)<\s*/%s\s*>)'%(dom_type, dom_type), html, re.S|re.M)
    for out_html, in_html in matched_doms:
        if '$$' in dom_template:
            new_dom_html = dom_template.replace('$$', in_html)
        else:
            new_dom_html = dom_template.replace('$', out_html)
        html = html.replace(out_html, new_dom_html)
    return html


def re_html_dom(html, rules):
    # 对_re_html_dom的增强，可以同时处理多个rules
    if not isinstance(rules, (list, tuple)):
        return html
    rules = rules[:100] # 最多进行100次的替换...不然性能实在太糟糕
    if len(rules) == 2: # 可能本身就是一个单一的 rule
        html = _re_html_dom(html, rules)
    for rule in rules:
        if not isinstance(rule, (list, tuple)):
            continue
        html = _re_html_dom(html, rule)
    return html


#coding: utf8
from __future__ import absolute_import
import os
from flask import request, g
from farbox_lite.utils import same_slash
from farbox_lite.compiler.site import get_site_doc, update_site_doc
from farbox_lite.utils.lazy import get_value_from_data


sites = {}


# webserver 运行之前, 进行的构建
def create_site(site_folder, site_folder_name=''):
    site_folder = same_slash(site_folder)
    site_doc = get_site_doc(site_folder)
    if not site_doc:
        return {}# ignore

    domains = get_value_from_data(site_doc, 'domains', [])
    if not isinstance(domains, (list, tuple)):
        domains = []
    for domain in domains:
        domain = domain.strip().lower()
        domain = domain.split('://')[-1]
        domain = domain.strip('/').split('/')[0]
        if '.' not in domain and domain not in ['localhost']:
            break
        key =  domain
        sites[key] = site_doc

    if site_folder_name and site_folder_name not in sites:
        sites[site_folder_name] = site_doc

    return sites



def create_sites(site_folders_root):
    # 一个 folder 左右所有 sites 的集合
    if not os.path.isdir(site_folders_root):
        return
    sub_names = os.listdir(site_folders_root)
    for sub_name in sub_names:
        if sub_name.startswith('.') or sub_name.startswith('~'):
            continue
        site_path = os.path.join(site_folders_root, sub_name)
        site_folder_name = sub_name.lower().strip()
        if os.path.isdir(site_path):
            create_site(site_path, site_folder_name=site_folder_name)


def get_site_from_domain(domain):
    if ':' in domain:
        domain = domain.split(':')[0]
    domain = domain.lower()

    # os.environ first!
    site_folder_name = os.environ.get('farbox_lite_site_folder_name')
    if site_folder_name:
        matched_site_doc = sites.get(site_folder_name.lower().strip())
        if matched_site_doc:
            return matched_site_doc

    site_doc = sites.get(domain) or sites.get(domain.replace('www.', '')) or sites.get('www.'+domain)
    if site_doc:
        return site_doc

    # 仍然没有, 则默认 hit 一个
    for k, site_doc in sites.items():
        if site_doc.get('template_resources'):
            return site_doc



def get_site_from_request():
    g_site = getattr(g, 'site', None)
    if g_site:
        site = g_site
    else:
        site = get_site_from_domain(request.host)
        g.site = site
    return site


def get_sites():
    return sites







#coding: utf8
from scss import Scss
from scss.errors import SassSyntaxError
compiler = Scss()
import os, re

def compile_css(content, filename=None):
    raw_content = content
    if filename and isinstance(filename, (str, unicode)):
        name, ext = os.path.splitext(filename)
        ext = ext.lower()
        if ext in ['.less']:
            content = re.sub(r'@(\w)', '$\g<1>', content)
            content = re.sub(r'\$media ', '@media ', content)
    # remove import tag
    content = re.sub(r'@import.*?[\r\n]', '', content, re.I)
    try:
        to_return = compiler.compile(content)
    except SassSyntaxError, e:
        try:
            message = str(e)
            to_return = '/*%s*/\n%s' % (message, raw_content)
        except:
            to_return = raw_content
    except Exception, e:
        to_return = raw_content

    return to_return




def compile_css_file(filepath):
    if not filepath:
        return
    if not os.path.isfile(filepath):
        return
    with open(filepath, 'rb') as f:
        raw_content = f.read()
    css_content = compile_css(raw_content, filename=os.path.split(filepath)[-1])
    if isinstance(css_content, unicode):
        css_content = css_content.encode('utf8')
    prefix, ext = os.path.splitext(filepath)
    css_filepath = prefix + '.css'
    with open(css_filepath, 'wb') as f:
        f.write(css_content)


# coding: utf8
from __future__ import absolute_import
from jinja2.environment import Environment
from jinja2.sandbox import SandboxedEnvironment
from jinja2.exceptions import UndefinedError, TemplateNotFound, SecurityError
from jinja2.runtime import Undefined
from werkzeug.exceptions import HTTPException
from flask import abort
import datetime, inspect
import types

from farbox_lite.template_system.func import get_functions_in_a_file_obj, AttrFunc
from farbox_lite.utils.functional import curry

from farbox_lite.template_system.functions import attr_patches
from farbox_lite.template_system.functions.attr_patches.models.date import Date
from farbox_lite.template_system.functions.attr_patches.models.text import Text

from farbox_lite.template_system.utils import return_error_message
from farbox_lite.utils.error import capture_error



all_functions = get_functions_in_a_file_obj(attr_patches.all_functions)
array_functions = get_functions_in_a_file_obj(attr_patches.array_functions)
dict_functions = get_functions_in_a_file_obj(attr_patches.dict_functions)
str_functions = get_functions_in_a_file_obj(attr_patches.str_functions)
date_functions = get_functions_in_a_file_obj(attr_patches.date_functions)

# 特定类型的数据，如果请求其 attr，可以找到对应的functions 进行处理
# 用isinstance(obj, (1,2) or 1) 来判断的，所以是以下的 match table 格式
func_match_rules = [
    [(list, tuple), array_functions],
    [dict, dict_functions],
    ((str, unicode), str_functions),
    [datetime.datetime, date_functions],
]



def render_attr_func_without_call(func): # h.xxx 这样的函数可以直接当做属性处理
    if not hasattr(func, '__call__') or not func:
        return func
    if getattr(func, 'is_attr_func', False):
        return func
    new_func = AttrFunc(func)
    return new_func



def get_attr_func(obj, attr, functions):
    # 在一个指定的functions 的 dict 中，试图取得obj.attr 时，返回对应的 function（如果有的话）
    first_arg = None # 被调用的时候隐含的参数，比如posts.sort_by_date, 其中实际调用的是posts.sort 这个属性
    original_func = None
    if attr in functions:
        original_func = functions[attr]
    elif '_by_' in attr: # posts.sort_by_date('-')
        name, first_arg = attr.split('_by_', 1)
        if name in functions:
            original_func = functions[name]
    if original_func:
        wrapped_func = curry(original_func, obj)
        # 不能被 wrap 的函数
        if hasattr(original_func, 'arg_spec'):
            # 比如返回一个 dict(如 humech)，就不能是一个 function 了，或者指定了就是一个属性
            args_length = len(original_func.arg_spec.args)
            if args_length == 1 and not original_func.arg_spec.varargs and not original_func.arg_spec.keywords:
                # 原函数仅仅接收一个参数
                return wrapped_func()
            elif not args_length:
                # 不接受任何参数的， curry 都是不行的, 其实就是相当于一个属性（的函数运行）
                return original_func()
        return wrapped_func


old_getattr_func = Environment.getattr
sandbox_getattr_func = SandboxedEnvironment.getattr
allowed_fields = ['_id', ]
def env_getattr(self, obj, attr, sandbox=False):

    # 是否是 API 本提供的属性
    is_namespace_field = False
    if attr.startswith('_') and '__' not in attr and not isinstance(obj, dict) and attr not in allowed_fields:
        try:
            obj_name = obj.__class__.__module__
            if 'farbox_lite.template_system.functions.namespace.' in obj_name:
                is_namespace_field = True
        except:
            pass

    # 字符串其实比较常用，避免过度解析
    # tryText是 attr 以_结尾 或者 attr 本身的名称为 content

    if attr in allowed_fields:
        getattr_func = old_getattr_func
    elif isinstance(obj, dict) and attr in obj:
        getattr_func = old_getattr_func
    elif is_namespace_field:
        getattr_func = old_getattr_func
    elif sandbox:
        getattr_func = sandbox_getattr_func
    else:
        getattr_func = old_getattr_func

    if attr.startswith('__'):
        return ''


    if attr.endswith('_'):
        try_Text = True
        attr = attr[:-1]
    else:
        if attr in ['content']:
            try_Text = True
        else:
            try_Text = False

    #if isinstance(obj, (str, unicode)): # obj 本身就是字符串的情况
    #    _obj = Text(obj)
    #    if hasattr(_obj, attr): # 比如 '123'.int
    #        return getattr(_obj, attr)
    #    else: # 比如 'abc'.title
    #        old_getattr_func_value = getattr_func(self, obj, attr)
    #        if not isinstance(old_getattr_func_value, Undefined):
    #            return old_getattr_func_value

    try_patch = False
    try:
        value = getattr_func(self, obj, attr)
        if isinstance(value, Undefined):
            try_patch = True
        elif hasattr(value, '__call__'):
            try_patch = True
    except UndefinedError:
        value = '' # by default
        try_patch = True

    # 这里应该触发错误, 如果不是UndefinedError 的话

    if isinstance(value, Undefined):
        try_patch = True

    # 使用数据类型的 class 直接进行处理
    if isinstance(value, datetime.datetime):
        return Date(value)
    elif isinstance(value, (str, unicode)) and try_Text:
        return Text(value, attr=attr, parent=obj)

    # 必须被 patch 的一些属性
    if not try_patch and isinstance(obj, (Text, Date)) and not hasattr(obj, attr):
        try_patch = True

    if try_patch:
        # 出现 UndefinedError 或者本身就是 Undefined， 那么放心大胆的 patch
        # 另外 函数的patch 优先级高于原生的，比如insert 对于一个 array
        all_matched_func = get_attr_func(obj, attr, all_functions)
        if all_matched_func:
            return all_matched_func
        for func_types, functions in func_match_rules:
            if isinstance(obj, func_types):
                matched_func = get_attr_func(obj, attr, functions)
                if matched_func is not None:
                    return matched_func

    if hasattr(value, '__call__') and value and hasattr(obj, '__class__') and \
            isinstance(value, (types.FunctionType, types.MethodType)):
        # 只有默认变量的一些子属性才有这样的功能
        obj_class = obj.__class__
        if hasattr(obj_class, '__module__'):
            obj_module = obj_class.__module__
            if isinstance(obj_module, (str, unicode)) and obj_module.startswith('farbox_lite.template_system.functions.namespace.'):
                value = render_attr_func_without_call(value)

    if value is None:
        value = ''

    return value

# patch
def env_call(self, context, context_obj, *args, **kwargs):
    # 实际上是要求运行obj(这个函数）, 默认变量都是进行 call的，以保证 jinja2 解析后的代码执行次序
    if not context_obj: # maybe Undefined
        return context_obj
    if not hasattr(context_obj, '__call__') and not args and not kwargs:
        return context_obj
    if getattr(context_obj, 'was_called', False):
        # 比如 site = site or something, 会导致 site 被多次 call...
        return context_obj

    # 不能被 call 的，本身就是函数，需要在模板里被调用
    # 因为所有变量都会被 call，在赋值的时候；而函数类型的，是后续要调用的，直接被 call 的话，获得类型就不对了，不再是函数，就没有办法在 api 里调用了。
    # env_call 跟 get_attr 又不一样，它更多是一级的变量调用，get_attr 是获取属性
    # context_obj_name = getattr(context_obj, '__name__', None)
    # (args, varargs, keywords, defaults). args is a list of the argument names (it may contain nested lists). varargs and keywords are the names of the * and ** arguments or None
    if not args and not kwargs and hasattr(context_obj, 'arg_spec') and context_obj.arg_spec:
        if context_obj.arg_spec.args or context_obj.arg_spec.varargs or context_obj.arg_spec.keywords:
            return context_obj # like i18n

    try:
        result = context.call(context_obj, *args, **kwargs)
        if not hasattr(result, '__call__'):
            try: setattr(result, 'was_called', True) # 一些变量
            except: pass
        return result
    except HTTPException, e:
        raise e
    except Exception, e:
        if isinstance(context_obj, (unicode, str, list, tuple, int, float)): # 基础数据类型，call的时候，返回其本身
            return context_obj
        error_info = "%s not a valid function, can't be called, error_info: %s" % (getattr(context_obj, 'func_name', ''), e.message)
        return return_error_message(error_info)


old_env_handle_exception = Environment.handle_exception
def env_handle_exception(self, exc_info=None, rendered=False, source_hint=None):
    # source_hint 是编译后的 jinja2 模板
    source_hint = source_hint or ''
    if not rendered:
        error = exc_info[1]
        filename = error.filename or ''
        if filename.endswith('.jade'):
            filename = '<a href="/service/template_code/%s">compiled %s</a>, ' \
                       'the line number may be different to your own code' % (filename, filename)
        error_information = "template file: %s \n line number:%s \n error message: %s" % (
            filename,
            error.lineno,
            error.message
        )
        error_info = dict(lineno=error.lineno, source=source_hint, message=error_information, debug=True)
        abort(500, error_info)
    elif exc_info[0] in [TemplateNotFound]:
        error_information = '%s TemplateNotFound' % exc_info[1].message
        abort(500, error_information)
    elif exc_info[0] in [UndefinedError, ValueError, TypeError]:
        capture_error()
        error_information = exc_info[1].message
        abort(500, error_information)
    else:
        # 这里不要尝试直接 500， 有可能是 redirect 或者其它错误
        return old_env_handle_exception(self, exc_info, rendered, source_hint)

Environment.getattr = env_getattr
Environment.call = env_call
Environment.handle_exception = env_handle_exception



#coding: utf8
from __future__ import absolute_import
from farbox_lite.compiler.utils import get_file_list
from farbox_lite.utils import get_relative_path, same_slash, smart_unicode
from farbox_lite.utils.functional import curry
import os



template_resources_cache = {}



def is_allowed_template_resource(root, filepath):
    if not os.path.exists(filepath):
        return False
    relative_path = get_relative_path(filepath, root=root)
    relative_path = relative_path.lower()
    if relative_path in ['data']:
        return False
    elif relative_path.startswith('data/'):
        return False
    if os.path.isfile(filepath):
        ext = os.path.splitext(filepath)[-1].lower().strip('.')
        if ext not in ['html', 'js', 'css', 'otf', 'eot', 'ttf', 'woff', 'woff2', 'png', 'jpg', 'gif']:
            return False
        else:
            return True
    elif os.path.isdir(filepath):
        return True
    else:
        return False


def raw_get_template_resources(root):
    root = same_slash(root)
    root = os.path.join(root, 'template') # templates' folder
    if not os.path.isdir(root):
        return {}
    template_resources = {}
    template_resource_mtimes = {}
    allowed_func = curry(is_allowed_template_resource, root)
    template_resource_filepaths = get_file_list(root, filter_func=allowed_func)
    for filepath in template_resource_filepaths:
        mtime = os.path.getmtime(filepath)
        relative_path = get_relative_path(filepath, root=root).strip().lower()
        if not os.path.isfile(filepath):
            continue
        with open(filepath, 'rb') as f:
            file_content = f.read()
        template_resources[relative_path] = file_content
        template_resource_mtimes[relative_path] = (filepath, mtime)

        # +号衍生的路由
        relative_parent, filename = os.path.split(relative_path)
        just_name, ext = os.path.splitext(filename)
        if '+' in just_name:
            for sub_name in just_name.strip().split('+'):
                sub_name = sub_name.strip()
                if not sub_name:
                    continue
                sub_path = '%s/%s%s' % (relative_parent, sub_name, ext)
                sub_path = sub_path.lstrip('/')
                template_resources[sub_path] = file_content
                template_resource_mtimes[sub_path] = (filepath, mtime)

    template_resources['_'] = template_resource_mtimes

    return template_resources


def get_template_resources(root):
    root = same_slash(root)
    if not os.path.isdir(root):
        return {}
    if root in template_resources_cache:
        template_resources = template_resources_cache.get(root)
        return template_resources
    template_resources = raw_get_template_resources(root)
    template_resources_cache[root] = template_resources
    return template_resources



def update_template_resources(root):
    template_resources = raw_get_template_resources(root)
    template_resources_cache[root] = template_resources
    return template_resources




# coding: utf8
from __future__ import absolute_import
from farbox_lite.template_system.func import get_functions_in_a_file_obj
default_variables = {}

sub_names = ['post', 'site', 'image', 'html', 'data',
             'request', 'response', 'shortcut', 'syntax_block'
             ]
for sub_name in sub_names:
    exec('from farbox_lite.template_system.functions.namespace import %s' % sub_name)
    exec('default_variables.update(get_functions_in_a_file_obj(%s))' % sub_name)

    try: # 非函数性质的
        exec('default_variables.update(%s.default_variables)' % sub_name)
    except:
        pass


# 实体的file class
parts = []
for part in parts:
    default_variables.update(get_functions_in_a_file_obj(part))






# coding: utf8
from __future__ import absolute_import

from farbox_lite.utils import smart_unicode
from farbox_lite.utils.functional import curry, cached_property
from farbox_lite.utils.lazy import get_value_from_data
from farbox_lite.render.apis.data import get_data, get_paginator, get_doc
from farbox_lite.template_system.utils import SafeUndefined

# category 相关的很难单独出来，所以放在 utils 里调用


def get_doc_category(doc):
    if isinstance(doc, dict) and doc.get('type') and doc.get('path'):
        doc_path = doc.get('path') or ''
        if '/' in doc_path:
            parent_path = doc_path.rsplit('/', 1)[0]
            return Category(parent_path)




class Category(object):
    def __init__(self, path_or_doc):
        if isinstance(path_or_doc, dict):
            self.raw = path_or_doc
            self.path = self.raw.get('path')
        else:
            path = path_or_doc
            path = smart_unicode(path).strip()
            folder = {}
            doc = get_doc(path=path)
            if doc and doc.get('type') == 'folder':
                folder = doc
            self.raw = folder
            self.path = path

    def __nonzero__(self):
        return bool(self.raw)


    def __getattr__(self, item):
        if item in self.raw:
            return self.raw.get(item)
        return self.__dict__.get(item, SafeUndefined())


    def __getitem__(self, item):
        return self.__getattr__(item)


    @cached_property
    def parents(self):
        # 得到所有的上级目录
        ps = []
        if not self.raw:
            return ps
        path_parts = self.path.split('/')[:-1][:50] # 最多50个层级
        parent_paths = []
        for i in range(len(path_parts)):
            parent_paths.append('/'.join(path_parts[:i+1]))
        parent_paths.reverse() # just reverse it for human
        parent_categories = []
        for parent_path in parent_paths:
            parent_category = Category(parent_path)
            if parent_category:
                parent_categories.append(parent_category)
        return parent_categories


    @cached_property
    def url(self):
        return u'/category/%s' % self.path



    @cached_property
    def posts(self):
        # 仅仅当前
        pager_name = '%s_child_posts' % self.path
        return get_data(type='post', path=self.path, pager_name=pager_name, level=1)

    @cached_property
    def posts_pager(self):
        pager_name = '%s_child_posts' % self.path
        return get_paginator(pager_name)


    @cached_property
    def all_posts(self):
        # 遍历所有
        pager_name = '%s_all_posts' % self.path
        return get_data(type='post', path=self.path, pager_name=pager_name)


    @cached_property
    def all_posts_pager(self):
        pager_name = '%s_all_posts' % self.path
        return get_paginator(pager_name)


    @cached_property
    def cover_url(self):
        return self.raw.get('cover') or ''
        
    @cached_property
    def cover(self):
        return self.cover_url

    @cached_property
    def images(self):
        # 仅仅当前
        pager_name = '%s_child_images' % self.path
        return get_data(type='image', path=self.path, pager_name=pager_name, level=1)


    @cached_property
    def images_pager(self):
        pager_name = '%s_child_images' % self.path
        return get_paginator(pager_name)


    @cached_property
    def all_images(self):
        # 遍历所有
        pager_name = '%s_all_images' % self.path
        return get_data(type='image', path=self.path, pager_name=pager_name)


    @cached_property
    def all_images_pager(self):
        pager_name = '%s_all_images' % self.path
        return get_paginator(pager_name)





#coding: utf8
from __future__ import absolute_import
from math import ceil
import flask
import urllib
from flask import request
from farbox_lite.utils.functional import cached_property
from farbox_lite.utils.lazy import to_int
from farbox_lite.utils import smart_unicode


def compute_auto_pages(pages, current_page=1, max_count=10):
    # 1 ...... n
    if max_count < 6:
        max_count = 6
    if pages <= max_count:
        return range(1, pages+1)

    just_head = range(1, max_count-1) + [0, pages]
    just_foot = [1, 0] + range(pages-max_count+1, pages+1)

    if current_page in just_head and current_page!=pages:
        if current_page < max_count/2:
            return just_head
    if current_page in just_foot and current_page!=1:
        if current_page > pages-max_count/2:
            return just_foot

    auto_fix_count = (max_count - 2*2)/2
    head = [1, 2]
    foot = [pages-1, pages]
    _middle_list = range(current_page-auto_fix_count+1, current_page+auto_fix_count)
    middle_list = []
    for i in _middle_list:
        if 1<i<pages and i not in head and i not in foot:
            middle_list.append(i)
    result =  head
    if result[-1]+1 not in middle_list:
        result.append(0) # head fill
    result += middle_list
    if result[-1] and result[-1]+1 not in foot:
        result.append(0)
    result += foot
    return result



class Paginator(object):
    def __init__(self, query, per_page=3, page=1, is_safe=False):
        # query is like db.entry.find({'something': 'something'})
        page = to_int(page, 1, 10000) # 最多不超过10k页
        per_page = to_int(per_page, 3, 1000) # 单页不超过1k条记录
        if page * per_page > 100000: # 数据总数不能超过10w条
            flask.abort(404)

        self.query = query
        self.per_page = per_page
        self.total_pages = int(ceil(float(self.total_count) / per_page))
        self.pages = self.total_pages # 总页码数
        self.page = page
        self.next_page = page + 1
        self.previous_page = page - 1
        self.is_safe = is_safe # 如果为真的，则不对输出结果进行关键字hidden

        self.default_max_page_numbers = 10

    def __getitem__(self, item):
        # 主要是做 for 循环用的
        if hasattr(self.list_object, '__getitem__'):
            return self.list_object.__getitem__(item)

    @cached_property
    def pre_page(self):
        return self.previous_page

    @cached_property
    def total_count(self):
        if isinstance(self.query, (list, tuple)):
            return len(self.query)
        if hasattr(self.query, 'count') and hasattr(self.query.count, '__call__'):
            return self.query.count()
        else:
            return len(self.query)

    @cached_property
    def has_previous(self):
        return self.page > 1

    @cached_property
    def has_pre(self):
        return self.has_previous

    @cached_property
    def has_next(self):
        return self.page < self.total_pages


    @staticmethod
    def get_page_url(page_number):
        if '/page/' in request.url:
            base = request.path.split('/page/',1)[0]
        else:
            base = request.path
        if page_number != 1:
            url = ('%s/page/%s'%(base, page_number)).replace('//','/')
        else:
            url = base.replace('//','/') or '/'
        if request.query_string:
            query_string = request.query_string
            if '%' in query_string:
                query_string = urllib.unquote(query_string)
            url += '?%s'%smart_unicode(query_string)
        url = url.replace('"', '%22').replace("'", '%27') # 避免被跨站
        return url

    @cached_property
    def previous_page_url(self):
        if self.has_previous:
            return self.get_page_url(self.previous_page)
        else:
            return '#'

    @cached_property
    def pre_page_url(self):
        return self.previous_page_url

    @cached_property
    def pre_url(self):
        return self.pre_page_url

    @cached_property
    def next_page_url(self):
        if self.has_next:
            return self.get_page_url(self.next_page)
        else:
            return '#'

    @cached_property
    def next_url(self):
        return self.next_page_url

    @cached_property
    def list_object(self):
        if self.page > self.total_pages or self.page < 1:
            return []
        start_at = (self.page-1) * self.per_page
        data = self.query[start_at: start_at+self.per_page]
        return list(data)

    @cached_property
    def page_numbers(self):
        return self.get_page_numbers()

    def set_default_max_page_numbers(self, numbers):
        # 设定 auto_pages 计算时，最大的长度跨度
        numbers = to_int(numbers)
        if isinstance(numbers, int) and 50>numbers>3:
            self.default_max_page_numbers = numbers
        return ''


    def get_page_numbers(self, max_count=None):
        if not max_count:
            max_count = self.default_max_page_numbers
        return compute_auto_pages(self.total_pages, current_page=self.page, max_count=max_count)



#coding: utf8
from __future__ import absolute_import
import os, datetime
from farbox_lite.compiler.posts.post import filepath_to_post_doc
from farbox_lite.compiler.utils import get_index_doc_filepath
from farbox_lite.utils import same_slash, get_relative_path


category_docs_cache = {}

def raw_get_category_doc(category_path, root=None):
    category_path = same_slash(category_path)
    index_doc_filepath = get_index_doc_filepath(category_path)
    relative_path = get_relative_path(category_path, root=root)
    path = relative_path.strip().lower()
    if index_doc_filepath and os.path.isfile(index_doc_filepath):
        category_doc = filepath_to_post_doc(index_doc_filepath, root=root)
    else:
        category_doc = {}
        category_doc['title'] = os.path.split(category_path)[-1]
    category_doc['type'] = 'folder'
    category_doc['path'] = path
    category_doc['filepath'] = category_path
    category_doc['url'] = '/category/%s' % path
    category_doc['date'] = datetime.datetime.fromtimestamp(os.path.getmtime(category_path))
    return category_doc




def get_category_doc(category_path, root=None):
    category_path = same_slash(category_path)
    if not root:
        root = os.path.dirname(category_path)
    if category_path in category_docs_cache:
        return category_docs_cache.get(category_path)
    else:
        category_doc = raw_get_category_doc(category_path, root=root)
        category_docs_cache[category_path] = category_doc
        return category_doc


def update_category_doc(category_path):
    category_doc = raw_get_category_doc(category_path)
    category_docs_cache[category_path] = category_doc





#coding: utf8
import datetime, re
from dateutil.parser import parse
import calendar
import time
from flask import g
from .lazy import get_value_from_data, to_float

def date_parse(timestr, parserinfo=None, utc_offset=8, **kwargs): # 分析日期的
    if utc_offset is None: # utc_offset=None， 表示是自动的，尝试从site里获得configs.utc_offset
        try:
            utc_offset = to_float(get_value_from_data(g, 'site.configs.utc_offset', 8))
            if utc_offset>13 or utc_offset<-12: # 要在一个合理的范围内
                utc_offset = 8
        except:
            utc_offset = 8
    if isinstance(utc_offset, (str,unicode)):
        utc_offset = to_float(utc_offset, 8)


    if not isinstance(timestr, (unicode, str)): #非字符串的，不处理
        #if hasattr(timestr, 'tzinfo') and timestr.tzinfo is None:  # 时间格式且无时区
        #    timestr -= datetime.timedelta(0, utc_offset*3600) # 调整时区偏差
        # 不需要调整时区，但这种直接传入的date，需确保是utcnow()获取的
        return timestr
    else:
        timestr = timestr.strip()
        if not timestr: # 空字符不处理
            return timestr

        # 网易邮箱的 Date...
        # Fri, 28 Apr 2017 20:09:08 +0800 (GMT+08:00)
        if ' +' in timestr:
            p1, p2 = timestr.split(' +', 1)
            timezone_str = p2.split(' ')[0]
            timestr = '%s +%s' % (p1, timezone_str)


    if isinstance(timestr, unicode):
        timestr = timestr.replace(u'\uff1a', ':')

    date = parse(timestr, parserinfo, **kwargs)

    date_offset_done = False

    if date.tzinfo is not None: # 保证数据有效性
        try:
            diff = date.utcoffset()
            date = date.replace(tzinfo=None)
            date = date - diff
            date_offset_done = True
        except ValueError: #比如exif中的信息有错，则不处理；但是mongodb又会转化，所以，要保证数据格式可用
            date = date.replace(tzinfo=None)

    # utc_offset now
    if date.tzinfo is None and not date_offset_done: # 如果不包含时区信息，则偏移，比如dropbox api提供的都是有时区信息的
        try:
            date -= datetime.timedelta(0, utc_offset*3600)
        except:
            # 囧, 可能会导致 date 的时间工差 <0 ....
            pass

    return date


def get_utc_timestamp():
    # equal to time.time
    #  function returns the number of seconds since the epoch as seconds in UTC.
    now = datetime.datetime.utcnow()
    return date_to_timestamp(now)


def get_isoformat(timestamp=None):
    '''Format the timestamp according to ISO8601 / RFC3339.'''
    if isinstance(timestamp, datetime.datetime):
        timestamp = date_to_timestamp(timestamp)
    timestamp = timestamp or get_utc_timestamp()
    t = time.gmtime(timestamp)
    return time.strftime('%Y-%m-%dT%H:%M:%SZ', t)

def get_rfcformat(timestamp=None):
    '''Format the timestamp according to RFC822.'''
    if isinstance(timestamp, datetime.datetime):
        timestamp = date_to_timestamp(timestamp)
    timestamp = timestamp or get_utc_timestamp()
    t = time.gmtime(timestamp)
    return time.strftime('%a, %d %b %Y %H:%M:%S GMT', t)


def date_to_timestamp(date, is_int=False):
    # date 应该是UTC时间
    if not isinstance(date, datetime.datetime):
        return date
    else: # 需要用calendar.timegm，不然会用本地的时区获得timestamp
        s = '%s.%s' % (int(calendar.timegm(date.timetuple())), date.microsecond) # 保留微秒
        if is_int:
            s = int(s)
        else:
            s = float(s)
        return s

def timestamp_to_date(timestamp, is_utc=True):
    if isinstance(timestamp, (str, unicode)):
        timestamp = timestamp.lower()
        if re.match(r'^\d+(\.\d+)?$', timestamp):
            timestamp = float(timestamp)
    if not isinstance(timestamp, (int, float)):
        # 数据类型不符合
        return timestamp
    if is_utc:
        return datetime.datetime.utcfromtimestamp(timestamp)
    else: # 当前计算机设定时区内的时间
        return datetime.datetime.fromtimestamp(timestamp)



# 获得图片的日期，exif信息中的优先
def get_image_date(exif, file_date=None, utc_offset=8):
    date = file_date
    try:
        exif_date = exif.get('datetime') or exif.get('datetime_mk')
        if not exif_date:
            return date
        exif_date = exif_date.replace(':', '-', 2)
        date_pat = re.search('([12][0-9]{3,3}-[0-9 :-]+)', exif_date)  # exif信息中可能夹杂了其它不干净的内容
        if date_pat:
            exif_date = date_pat.groups()[0]
        date = date_parse(exif_date, utc_offset=utc_offset)
    except:
        pass
    return date


def get_utc_today():
    return datetime.datetime.utcnow().strftime('%Y-%m-%d')


#coding: utf8
from __future__ import absolute_import
from flask import request
import time

def time_cost_handler(response=None):
    if not response: # before request
        request.environ['request_at'] = time.time()
    else: # after request
        request_at = request.environ.get('request_at')
        if request_at:
            time_cost = time.time() - request_at
            response.headers['x-render-time'] = str(time_cost) # 渲染消耗的时间，单位秒
        return response



#coding: utf8
from __future__ import absolute_import
import os
import operator
from farbox_markdown.util import is_a_markdown_file
from farbox_lite.compiler.utils import get_file_list
from farbox_lite.compiler.posts.post import filepath_to_post_doc


# 此处不处理缓存，在 API 的逻辑中, 有 update 的逻辑, 因为有 cache_uuid 的逻辑


def is_a_post_markdown_file(filepath):
    filename = os.path.split(filepath)[-1]
    if filename.lower().strip() in ['index.md', 'index.txt', 'index.mk', 'index.markdown']:
        # index as category
        return False
    if '/template/' in filepath:
        return False
    if not os.path.exists(filepath):
        return False
    elif os.path.isdir(filepath):
        return True
    else:
        return is_a_markdown_file(filepath)



def get_compiled_md_docs(root):
    md_docs = []
    md_filepaths = get_file_list(root, split=False, filter_func=is_a_post_markdown_file)
    for filepath in md_filepaths:
        doc = filepath_to_post_doc(filepath, root=root)
        if doc:
            md_docs.append(doc)
    md_docs = sorted(md_docs, key=operator.itemgetter('date'), reverse=True)
    return md_docs










# coding: utf8
from __future__ import absolute_import
from flask import g, request
import datetime
from jinja2.runtime import Undefined

from farbox_lite.render.apis.data import get_doc, has_doc
from farbox_lite.render.apis.site import get_site_from_request

from farbox_lite.utils.functional import cached_property
from farbox_lite.utils.web.cache_for_function import cache_result
from farbox_lite.template_system.functions.namespace.parts.social import get_social_items, get_social_links






class Site(object):
    def __init__(self):
        raw_site = get_site_from_request()
        raw_site = raw_site or {}

        self.raw = raw_site
        self.id = raw_site.get('id') or raw_site.get('_id') or ''



    def __getattr__(self, item):
        if item in self.raw:
            return self.raw.get(item)
        else:
            extra_data = self.raw.get('_', {})
            if item.startswith("_") and item.lstrip("_") in extra_data:
                return extra_data.get(item.lstrip("_"))
        return self.__dict__.get(item, Undefined())


    @cached_property
    def socials(self):
        socials_html = get_social_links()
        return socials_html

    @cached_property
    def social_items(self):
        return get_social_items()

    @cached_property
    def just_socials(self):
        return get_social_links()




@cache_result
def site():
    # 当前网站的数据对象
    return Site()



# coding: utf8
from functools import partial

class cached_property(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, type):
        res = instance.__dict__[self.func.__name__] = self.func(instance)
        return res

curry = partial




# coding: utf8
from __future__ import absolute_import
from flask import  g, request
from farbox_lite.utils.web.data import make_tree as _make_tree

from farbox_lite.render.apis.data import get_paginator as _get_paginator, get_doc as _get_doc,\
    get_data as _get_data, has_doc as _has_doc
from farbox_lite.utils.web.path import get_request_offset_path
from farbox_lite.utils.web.cache_for_function import cache_result



class Data(object):
    @staticmethod
    @cache_result
    def make_tree(docs, kept_fields=None):
        return _make_tree(docs, kept_fields)

    @staticmethod
    @cache_result
    def get_data(*args, **kwargs):
        return _get_data(*args, **kwargs)


    @staticmethod
    @cache_result
    def get_paginator(index_or_name=0):
        return _get_paginator(index_or_name)


    @staticmethod
    @cache_result
    def get_doc(path=None, match=True, **kwargs):
        # 有指定在某个目录下的，则补全 path 本身, 可以使用 url其实
        path = path or ''
        root = kwargs.pop('root', None) or ''
        if root and isinstance(root, (str, unicode)) and len(root) < 100:
            path = '/%s/%s' % (root.strip('/'), path.strip('/'))
            path = path.replace('//', '/') # 两个 // 是不允许的

        doc = _get_doc(path, match=match, **kwargs)
        if not doc:
            return doc
        as_context_doc = kwargs.get('as_context_doc')
        if as_context_doc and doc:
            g.doc = doc
        return doc

    def doc(self):
        # 返回纯粹的一个 doc 对象，比如一个 csv 类型的
        offset_path = get_request_offset_path(1)
        _doc = _get_doc(offset_path)
        if not _doc:
            _doc = _get_doc(request.path)
        return _doc

    @staticmethod
    def has(path): # , must_doc=False
        return _has_doc(path)



# 这种固定变量空间，要慎重使用 cached_property
#default_variables = {
#    'h': Html(),
#}

@cache_result
def d():
    return Data()

@cache_result
def data():
    return Data()



#coding: utf8
from __future__ import absolute_import
from .files import get_file_docs, get_cache_uuid as get_files_cache_uuid
from .posts import get_post_docs, get_cache_uuid as get_posts_cache_uuid
from .utils import sort_docs_by_field, cached, get_sub_docs_by_field, get_sub_docs_by_path
import os
from farbox_lite.compiler.files import filepath_to_file_doc


def get_cache_uuid():
    cache_uuid = '%s-%s' % (get_files_cache_uuid(), get_posts_cache_uuid())
    return cache_uuid

cache_wrapper = cached(cache_key_func=get_cache_uuid)


@cache_wrapper
def get_all_docs(root):
    post_docs = get_post_docs(root)
    docs = post_docs + get_file_docs(root)
    category_docs = []
    category_filepaths = set()
    posts_count_db = {}
    for doc in post_docs:
        filepath = doc.get('filepath')
        if filepath:
            category_filepath = os.path.dirname(filepath)
            category_filepaths.add(category_filepath)
            posts_count_db[category_filepath] = posts_count_db.get(category_filepath, 0)+1
    for category_filepath in category_filepaths:
        if category_filepath == root:
            continue
        category_doc = filepath_to_file_doc(category_filepath, root=root)
        if category_doc:
            category_docs.append(category_doc)
            category_doc['posts_count'] = posts_count_db.get(category_filepath, 0)
    docs += category_docs
    docs = sort_docs_by_field(docs, field='-date')
    return docs


@cache_wrapper
def get_docs_map(root):
    docs_map  = {}
    docs = get_all_docs(root)
    for doc in docs:
        path = doc.get('path')
        if not path:
            continue
        docs_map[path] = doc
    return docs_map


@cache_wrapper
def get_all_docs_by_field(root, field, value, sort=None):
    all_docs = get_all_docs(root)
    docs = get_sub_docs_by_field(all_docs, field, value)
    docs = sort_docs_by_field(docs, field=sort)
    return docs


@cache_wrapper
def get_all_docs_by_types(root, types, sort=None):
    all_docs = get_all_docs(root)
    docs = []
    for doc_type in types:
        type_docs = get_sub_docs_by_field(all_docs, 'type', doc_type)
        if type_docs:
            docs += type_docs
    docs = sort_docs_by_field(docs, field=sort)
    return docs


def get_doc_by_path(path, root):
    relative_path = path.strip('/').strip().lower()
    docs_map = get_docs_map(root)
    doc = docs_map.get(relative_path)
    return doc



def has_doc(path, root):
    root = root.rstrip('/')
    relative_path = path.strip('/').strip()
    full_path = os.path.join(root, relative_path)
    if os.path.exists(full_path):
        return True
    else:
        return False

@cache_wrapper
def get_all_sub_docs_by_path(root, raw_docs, parent_path, is_direct=False):
    # 相当于 get_sub_docs_by_path 带 cache 的逻辑
    all_sub_docs = get_sub_docs_by_path(root, raw_docs, parent_path, is_direct=is_direct)
    return all_sub_docs


@cache_wrapper
def get_next_doc(current_doc, root, sort='-date', same_folder=False):
    if not current_doc or not isinstance(current_doc, dict):
        return None
    current_doc_filepath = current_doc.get('filepath')
    if not current_doc_filepath:
        return None
    doc_type = current_doc.get('type')
    if not doc_type:
        return None
    docs = get_all_docs_by_field(root, field='type', value=doc_type, sort=sort)
    if same_folder: # 位于当前同一个目录之内
        parent_path = os.path.dirname(current_doc_filepath)
        docs = get_sub_docs_by_path(root, docs, parent_path=parent_path, is_direct=True)
    hit_next = False
    for doc in docs:
        if doc.get('filepath') and doc.get('filepath') == current_doc.get('filepath'):
            hit_next = True
            continue
        if hit_next:
            return doc






# coding: utf8
from __future__ import absolute_import
from flask import g, abort, send_file
import io
import re
from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.lazy import to_int
from farbox_lite.utils.web.cache_for_function import cache_result
from farbox_lite.utils.web.response import force_redirect, force_response


class Response(object):
    def __init__(self):
        self.set_property_allowed = True


    def __setattr__(self, key, value):
        if key in ['type', 'content_type', 'mime_type', 'mimetype']:
            self.set_content_type(value)
        elif key in ['status_code', 'code', 'status']:
            self.set_status_code(value)
        # anyway
        self.__dict__[key] = value


    def __getattr__(self, item):
        if item in ['type', 'content_type', 'mime_type', 'mimetype']:
            return self.__dict__.get('type') or 'text/html'
        elif item in ['code', 'status', 'status_code']:
            return self.__dict__.get('code') or 200

    def set_content_type(self, content_type):
        if isinstance(content_type, basestring):
            content_type = smart_str(content_type)
            g.response_content_type = content_type
            self.__dict__['type'] = content_type

        return ''

    def set_status_code(self, code):
        code = to_int(code) or 200
        if code in [
            200, 201, 202, 203, 204, 205, 206,
            300, 301, 302, 303, 304, 305, 306, 307,
            400, 401, 402, 403, 404, 405, 406, 408, 409, 410]:
            g.response_code = code
            self.__dict__['code'] = code
        return ''

    @staticmethod
    def redirect(url, code=302, keep_site_id=True):
        force_redirect(url, code=code, keep_site_id=keep_site_id)


    @staticmethod
    def raise_404(description=''):
        if getattr(g, 'response', None):
            return 'not allowed because for other response existing'
        # 触发错误，由 render 这个函数处理错误，从而实现404页面
        if description:
            description = smart_unicode(description)
            g.error_description = description
        g.response_code = 404
        abort(404, description)

    @classmethod
    def raise404(cls, description=''):
        return cls.raise_404(description)

    def set_header(self, key, value):
        if isinstance(key, (str, unicode)) and len(key) < 50 and re.match('^[a-z0-9-_]+$', key, flags=re.I):
            value = smart_unicode(value)
            if not isinstance(getattr(g, 'user_response_headers', None), dict):
                g.user_response_headers = {}
            g.user_response_headers[key] = value
            return ''
        else:
            return 'this key is not allowed'

    def response(self, content, mime_type='text/html', as_file=False, **kwargs):
        # 相当于一个断点，直接拦截当前的 response 进行新的返回
        if as_file: # 作为附件发出
            file_response = send_file(io.BytesIO(content), mimetype=mime_type, **kwargs)
            content = file_response
        return force_response(content, mime_type)


    def set_cookie(self, key, value):
        if not hasattr(g, 'cookies'):
            g.cookies = {}
        g.cookies[key] = value
        return ''

    def delete_cookie(self, key):
        if not hasattr(g, 'cookies_to_delete'):
            g.cookies_to_delete = []
        if key not in g.cookies_to_delete:
            g.cookies_to_delete.append(key)
        return ''





@cache_result
def response():
    return Response()


#coding: utf8
from flask import request, g, redirect, Response, abort, send_file, current_app, make_response
import os
import ujson as json

from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.mime import guess_type
from farbox_lite.utils.url import join_url, get_get_var


def _handle_redirect_url(url, keep_site_id=True):
    if url.startswith('?'):
        url = request.path + url
    protocol = request.environ.get('HTTP_X_PROTOCOL') or 'http'
    if protocol == 'https':
        if not url.startswith('http://') and not url.startswith('https://'):
            # 相对路径
            url = 'https://%s/%s' % (request.host, url.lstrip('/'))
        elif url.startswith('http://'):
            # 绝对路径
            url = url.replace('http://', 'https://', 1)

    if keep_site_id:
        # 保留 site_id
        if request.args.get('site_id'):
            if 'site_id=' not in url:
                url = join_url(url, site_id=request.args.get('site_id'))
        elif request.referrer:
            site_id_in_referrer = get_get_var(request.referrer, 'site_id')
            if site_id_in_referrer:
                if 'site_id=' not in url:
                    url = join_url(url, site_id=site_id_in_referrer)

    return url



def jsonify(data):
    try:
        data = json.dumps(data)
    except:
        data = json.dumps(dict(error='json_error'))
    response = Response(data, mimetype='application/json')
    return response


def send_plain_text(text):
    response = Response(text, mimetype='text/plain')
    return response


def p_redirect(url, *args, **kwargs):
    # 保证从https的url中跳转也是https的
    url = _handle_redirect_url(url)
    return redirect(url, *args, **kwargs)


def force_redirect(url, *args, **kwargs):
    if not isinstance(url, (str, unicode)):
        return ''
    keep_site_id = kwargs.pop('keep_site_id', True)
    url = _handle_redirect_url(url, keep_site_id)
    # 通过error 410的跳转来实现
    if not args and not kwargs: # just url
        abort(410, url)
    redirect_s = json.dumps(dict(url=url, args=args, kwargs=kwargs))
    abort(410, redirect_s)


def force_response(content, mime_type='text/html'):
    # 相当于一个断点，直接拦截当前的 response 进行新的返回
    if isinstance(content, Response): # 已经是 response 了
        abort(422, content)
    if isinstance(content, (dict, list, tuple)):
        try:
            content = json.dumps(content)
            if not mime_type or mime_type == 'text/html':
                mime_type = 'application/json'
        except:
            content = smart_unicode(content)
    else:
        content = smart_unicode(content)
        force_response_pre_handler = getattr(g, 'force_response_pre_handler', None)
        if force_response_pre_handler and hasattr(force_response_pre_handler, '__call__'):
            content = force_response_pre_handler(content)
    if isinstance(content, (str, unicode)):
        content = make_response(content)
    if isinstance(content, Response) and isinstance(mime_type, (str, unicode)):
        content.content_type = mime_type
    abort(422, content)



ERROR_MESSAGES = {
    503: 'Wait for a Moment, and Try Again.',
    400: 'Bad Request',
    401: 'You Need to Login Now.',
    404: 'Page not Found or this Request is not Allowed',
}

def json_if_error(code, message=''):
    message = message or ERROR_MESSAGES.get(code, '')
    result = dict(error_code=code, message=message)
    response = make_response(jsonify(result))
    response.status_code = code
    return response



def is_doc_modified(doc, date_field='date'):
    if not doc:
        return True
    #if request.environ.get('HTTP_PRAGMA') in ['no-cache']:
    #    return True
    if request.environ.get('HTTP_CACHE_CONTROL') in ['no-cache']:
        return True
    if date_field in doc:
        date = smart_str(doc[date_field])
        date_in_request = request.environ.get('HTTP_IF_MODIFIED_SINCE')
        if date_in_request == date: # and request.cache_control.max_age
            return False
    return True


def get_304_response():
    # 仅仅给出一个空 response
    response = make_response('')
    response.status_code = 304
    return response


def get_status_response(error_info='', status_code=404):
    error_info = error_info or 'not found'
    response = Response(error_info)
    response.status_code = status_code
    return response

def set_304_response_for_doc(doc, response, date_field='date'):
    if date_field in doc:
        date = smart_str(doc[date_field])
        response.headers['Last-Modified'] = date
    return response







def is_filepath_modified(filepath):
    try:
        mtime = int(os.path.getmtime(filepath))
        if request.environ.get('HTTP_CACHE_CONTROL') in ['no-cache']:
            return True
        date_in_request = request.environ.get('HTTP_IF_MODIFIED_SINCE')
        try:
            date_in_request = int(date_in_request)
            if date_in_request == mtime:
                return False
        except:
            pass
    except:
        pass
    # at last
    return True



def send_file_with_304(filepath):
    if not os.path.isfile(filepath):
        return # ignore
    if not is_filepath_modified(filepath):
        return get_304_response()
    else:
        response = send_file(filepath, mimetype=guess_type(filepath, 'application/octet-stream'))
        try:
            mtime = int(os.path.getmtime(filepath))
            response.headers['Last-Modified'] = smart_str(mtime)
        except:
            pass
        return response




def add_response_header(k, v):
    if not hasattr(g, 'more_headers'):
        g.more_headers = {}
    g.more_headers[k] = v




# coding: utf8
from jinja2.environment import Environment
from .attr import *
from .jinja_context import *
#from farbox_lite.template_system.env import FarboxLiteEnvironment


# 这个是在外部被调用，从而被 patch 的



#coding: utf8
from __future__ import absolute_import

paginator = u"""{% set paginator_style = paginator_style or 'simple' %}

{% set paginator_css_class = 'paginator pager pagination' %}
{% if  not paginator.has_pre and not paginator.has_next %}
{% set paginator_css_class = paginator_css_class + " no_pages" %}

{% endif %}
<div class="{{paginator_css_class}}" >
  <div class="paginator_container pagination_container">{% if  paginator.has_pre or paginator.has_next %}{% if  paginator_style == 'auto' %}{% if  max_count %}{{ paginator.set_default_max_page_numbers(max_count) }}

{% endif %}
{% set omission_mark = omission_mark or '...' %}
{% for page_number in paginator.page_numbers %}{% if  page_number %}
{% set is_current = True if paginator.page==page_number else False %}
<a href="{{paginator.get_page_url(page_number)}}"  class="{{'current' if is_current else ''}}" >{{ page_number }}</a>

{% else %}<span class="fill">{{ omission_mark }}</span>

{% endif %}

{% endfor %}
{% elif  paginator_style == 'mini' %}
{% set max_count = max_count or 5 %}
{% if  max_count < 3 %}
{% set max_count = 3 %}

{% endif %}{% if  max_count > 20 %}
{% set max_count = 20 %}

{% endif %}
{% set middle_count = (max_count/2).int %}
{% if  paginator.page <= middle_count %}
{% set page_begin = 1 %}

{% else %}
{% set page_begin = paginator.page - middle_count %}

{% endif %}
{% if  paginator.page + middle_count + 1 > paginator.total_pages %}
{% set page_end = paginator.total_pages + 1 %}

{% else %}
{% set page_end = paginator.page + middle_count + 1 %}

{% endif %}
{% if  paginator.has_pre %}<a href="{{paginator.pre_page_url}}"  rel="prev" class="prev">{{ pre_label }}</a>

{% endif %}{% for page in range(page_begin, page_end) %}
{% set a_class = 'page-number current' if page==paginator.page else 'page-number' %}
<a href="{{paginator.get_page_url(page)}}"  class="{{a_class}}" >{{ page }}</a>

{% endfor %}{% if  paginator.has_next %}<a href="{{paginator.next_page_url}}"  rel="next" class="next">{{ next_label }}</a>

{% endif %}
{% else %}{% if  paginator.has_pre %}<a href="{{paginator.pre_url}}"  class="btn pre newer-posts newer_posts">{{ pre_label }}</a>

{% endif %}{% if  show_page_of %}<span class="page_number page-number">{{ 'Page %s of %s' % (paginator.page, paginator.pages) }}</span>

{% endif %}{% if  paginator.has_next %}<a href="{{paginator.next_url}}"  class="btn next older-posts older_posts">{{ next_label }}</a>

{% endif %}
{% endif %}

    <div style="clear:both;height:0;"></div>

{% endif %}
  </div>

</div>"""


#coding: utf8
from __future__ import absolute_import
import re, urllib
from flask import request, g
from .url import get_url_without_prefix

def to_folder_path(path): # 左前是没有/的
    if type(path) in [list, tuple]:  # list to path (str)
        path = '/'.join(path)
    return path.strip('/') + '/'


def is_same_path(p1, p2):
    p1 = p1 or ''
    p2 = p2 or ''
    return p1.strip('/').lower() == p2.strip('/').lower()


def split_path(path):
    # 将一个path转为以/分割的parts，如果有包含~~的，则以~~为第一个split
    path = urllib.unquote(path)
    path = re.sub(r'/page/\d+/?$', '/', path) # 去掉/page/1这样的后缀

    if '~~' in path:
        # /blog/post/~this/is/post/path
        head, tail = path.split('~~', 1)
        parts = [head.strip('/')] + tail.split('/')
    else:
        parts = path.split('/')

    return filter(lambda x: x, parts)


def get_offset_path(path, offset=1):
    # /a/b/c/d 如果 offset=1, 则获得 b/c/d, 如果是2，则是 c/d
    if not offset or not isinstance(offset, int):
        return path
    else:
        raw_path = path
        path = path.lstrip('/')
        parts = split_path(path)
        path = '/'.join(parts[offset:])
        if raw_path.endswith('/'):
            path += '/'
        return path


def get_request_offset_path(offset=1, path=None):
    # 得到偏移几个单位的url，头没有  /
    if path is None:
        path = get_path()
    path = get_offset_path(path, offset)
    return path.lstrip('/')



def get_path():
    # 去掉/page/<int> 这样分页信息后的path, 并且不以/开头
    path_got = getattr(g, 'url_path', None)
    if path_got:
        return path_got
    # path = request.path.lower()
    path = request.path # 保留大小写，想获取 tags 的时候，需要区分大小写敏感的
    page_finder = re.compile(r'(.*?)/page/(\d+)/?$')
    page_result = page_finder.search(path)
    if page_result:
        path, page = page_result.groups()
        g.page = int(page) #存在g里面，这样后面供页面使用的函数，可以取这个值
    path = path.lstrip('/')
    g.url_path = path
    return path



def get_request_path():
    path = get_path()
    return '/%s' % path.lstrip('/')


def get_request_path_without_prefix():
    path = get_request_path()
    return get_url_without_prefix(path)


def get_request_offset_path_without_prefix(offset=1):
    path = get_request_path_without_prefix()
    return get_request_offset_path(offset, path=path)



def get_default_root_by_site_route():
    current_site_route = getattr(g, 'current_site_route', None)
    if current_site_route and isinstance(current_site_route, dict):
        default_root = current_site_route.get('route_root') or ''
        if default_root:
            return default_root
    return None


# coding: utf8
from __future__ import absolute_import
import re
from flask import g, request
from farbox_markdown.util import is_a_markdown_file

from farbox_lite.render.apis.data import get_data, get_paginator, get_doc
from farbox_lite.template_system.functions.namespace.utils.doc_list_utils import DocList
from farbox_lite.utils import smart_unicode
from farbox_lite.utils.functional import curry, cached_property
from farbox_lite.utils.web.path import get_request_offset_path_without_prefix
from farbox_lite.utils.web.cache_for_function import cache_result
from farbox_lite.utils.lazy import get_value_from_data, to_int


class Posts(DocList):
    def __init__(self):
        DocList.__init__(self, 'posts')
        self.status = 'public' # by default

    def __getattr__(self, item):
        # 找不到attribute的时候，会调用getattr
        if re.match('^recent_?\d+$', item): # 获得最近的几篇文章
            limit = re.search('\d+', item).group()
            limit = to_int(limit)
            return self.get_recent(limit)


    def __getitem__(self, item):
        if hasattr(self.list_obj, '__getitem__'):
            return self.list_obj.__getitem__(item)


    def __nonzero__(self):
        return bool(self.list_obj)

    def __iter__(self):
        # 返回一个迭代器，用于 for 的命令
        obj = self.list_obj
        self.pager_name = 'posts'
        if hasattr(obj, '__iter__'):
            return obj.__iter__()
        return obj


    @cached_property
    def length(self):
        return len(self.list_obj)


    def set_status(self, status):
        if isinstance(status, (str, unicode)):
            status = status[:35]
            self.status = status
        return ''


    def get_recent(self, limit=8):
        # 可以获得最近的几篇文章
        limit = to_int(limit, 8)
        limit = min(100, limit)
        return get_data(type='post', limit=limit, sort='-date', with_page=False, path=self.under)

    def get_recent_posts(self, limit=8):
        return self.get_recent(limit)


    def get_by_status(self, status='public', limit=None):
        # 按照 status 获得对应的文章
        if not isinstance(status, (int, str, unicode, float)):
            return []
        status = smart_unicode(status)
        self.pager_name = pager_name = '%s_posts' % status
        return get_data(types='post', status=status, pager_name=pager_name, min_limit=self.min_per_page, limit=limit)


    @property
    def pager(self):
        # 由于获得分页的数据对象
        # 比如可以直接调用  posts.pager....
        return get_paginator(self.pager_name)

    @property
    def paginator(self):
        return self.pager


    @property
    def all(self):
        self.pager_name = 'all_posts'
        return self._all


    @cached_property
    def _all(self):
        # 所有日志
        return get_data(types='post', pager_name='all_posts', min_limit=self.min_per_page)


    @cached_property
    def post(self):
        # 在 functions.namespace.shortcut 中，将 post 本身作为一个快捷调用，可以直接调用
        doc_in_g = getattr(g, 'doc', None)
        if doc_in_g and isinstance(doc_in_g, dict) and doc_in_g.get('type') == 'post':
            return doc_in_g

        # 得到当前 url 下对应的 post
        hide_post_prefix = get_value_from_data(g, 'site.configs.hide_post_prefix', default=False)
        if hide_post_prefix: # 不包含 /post/ 的 url
            url_path = self.request_path.lstrip('/')
        else: # 可能是/post/<url>的结构
            url_path = get_request_offset_path_without_prefix(offset=1)
        post_doc = get_data(type='post', url_path=url_path, return_one=True)

        if not post_doc and is_a_markdown_file(self.request_path): # 直接是markdown文件的路径
            post_doc = get_data(type='post', path=self.request_path, return_one=True)
            return post_doc # 避免写入g.doc todo why????

        if not post_doc and url_path and '/' in url_path: # sub path 的对应，offset 一次，让 markdown 本身作为 template 成为可能
            url_path = '/'.join(url_path.split('/')[:-1])
            post_doc = get_data(type='post', url_path=url_path, return_one=True)


        if post_doc: # 写入g.doc，作为上下文对象参数来处理
            g.doc = post_doc
        return post_doc


    @cached_property
    def list_obj(self):
        # 针对默认 的 url 规则，自动对应默认的 posts 的调用
        # 主要就是posts直接被 for 的逻辑调用了
        pager_name = 'posts'
        if self.keywords: # 关键词搜索放在最前面
            under = request.args.get('under') or self.under
            return get_data(type='post', keywords=request.args.get('s'), path=under,
                            limit=50, pager_name=pager_name, min_limit=self.min_per_page,
                            sort=self.sort or '-date', status=self.status)

        if self.request_path.startswith('/category/'): # 指定目录下的
            return get_data(type='post', path=get_request_offset_path_without_prefix(1), pager_name=pager_name, min_limit=self.min_per_page,
                            sort=self.sort, status=self.status)
        if self.request_path.startswith('/tags/') or self.request_path.startswith('/tag/'):
            _tags = get_request_offset_path_without_prefix(offset=1).strip('/')
            if _tags:
                _tags = [tag for tag in _tags.split('+') if tag]
                return get_data(type='post', tags=_tags, pager_name=pager_name, min_limit=self.min_per_page,
                                sort=self.sort, status=self.status)
        # at last
        path = self.under
        if not path and request.args.get('c'):
            # 相当于分类的路径传入
            path = request.args.get('c')
        _posts = get_data(type='post', path=path, pager_name=pager_name, min_limit=self.min_per_page, sort=self.sort, status=self.status)
        return _posts

    @cached_property
    def counts(self):
        self_list_obj = self.list_obj # 先行调用
        if self.pager:
            return self.pager.total_count
        else:
            return 0

    @cached_property
    def posts_count(self):
        return self.counts

    @cached_property
    def keywords(self):
        return request.args.get('s') or ''

    @cached_property
    def tags(self):
        _tags = get_request_offset_path_without_prefix(offset=1)
        if _tags == '/':
            return []
        else:
            _tags = [tag for tag in _tags.split('+') if tag]
            return _tags



@cache_result
def posts():
    return Posts()


#/usr/bin/env python
# coding: utf8
from setuptools import setup, find_packages
from farbox_lite import version

setup(
    name='farbox_lite',
    version=version,
    description='FarBox Lite Version',
    author='Hepochen',
    author_email='hepochen@gmail.com',
    include_package_data=True,
    packages=find_packages(),
    license = 'GPL',
    install_requires = [
        'setuptools',
        'farbox_markdown',
        'pygments',
        'shortuuid',
        'unidecode',
        'python-dateutil',
        'user_agents',
        'ujson',
        'flask==0.10',
        'jinja2==2.8',

    ],

    platforms = 'linux',
)


# coding: utf8
from __future__ import absolute_import
from werkzeug.exceptions import HTTPException
from jinja2.runtime import Undefined

from farbox_lite.utils import smart_unicode, UnicodeWithAttrs



def return_error_message(message):
    if 'error_ignored' in message:
        error_info = UnicodeWithAttrs('')
    else:
        error_info = UnicodeWithAttrs('<span style="color:red" class="template_api_error">&lt;%s&gt:</span>' % message)
    error_info.is_error = True
    return error_info



def run_function_safely(func):
    def _func(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except HTTPException, e:
            raise e
        except Exception, e:
            message = '`%s` usage error @ %s' % (getattr(func, 'func_name', 'function'), getattr(type(e), '__name__', ''))
            e_message = getattr(e, 'message', '')
            if e_message:
                message = '%s: %s' % (message, smart_unicode(e_message))
            return return_error_message(message)
    return _func






class SafeUndefined(Undefined):
    # 直接融入到用户的API渲染的页面中，我们自己程序中不做callback
    def __init__(self, *args, **kwargs):
        self._empty_list = []
        Undefined.__init__(self, *args, **kwargs)

    def __iter__(self):
        return self._empty_list.__iter__()

    def _fail_with_undefined_error(self, *args, **kwargs):
        return self

    def __getattr__(self, name):
        return self

    __add__ = \
    __radd__ = \
    __getitem__ = \
    __sub__ = \
    __mul__ = \
    __div__ = \
    __lt__ = \
    __le__ = \
    __gt__ = \
    __ge__ = _fail_with_undefined_error

    def __call__(self, *args, **kwargs):
        if self._undefined_name in ['caller', 'join', 'json', 'content']:
            return ''
        return return_error_message('`%s` is not a valid function' % self._undefined_name)


    def __repr__(self):
        return 'SafeUndefined'


# coding: utf8
from __future__ import absolute_import
from flask import request as _request, g
import copy, re, os, time

from farbox_lite.utils.functional import cached_property, curry
from farbox_lite.utils import smart_unicode
from farbox_lite.utils.lazy import to_int
from farbox_lite.utils.mime import guess_type
from farbox_lite.utils.url import join_url, is_same_url, decode_url_arg as _decode_url_arg, encode_url_arg as _encode_url_arg

from farbox_lite.utils.web.response import force_redirect
from farbox_lite.utils.web.request import get_visitor_ip, get_language, get_user_agent
from farbox_lite.utils.web.path import get_request_offset_path, get_request_offset_path_without_prefix, get_path
from farbox_lite.utils.web.url import get_url_with_prefix, get_url_without_prefix
from farbox_lite.utils.web.cache_for_function import cache_result





class Request(object):
    def __init__(self):
        self.url_fields = ['path', 'url', 'base_url', 'url_root', ]
        self.fields = ['form', 'args', 'values', 'method',
              'json', 'host', 'data', 'account_id', 'cookies']
        self.safe_fields = ['form', 'args', 'values', 'method', 'json', 'data', 'referrer']
        self.keep_url_safe = getattr(g, 'keep_request_url_safe', False)
        self.__setattr__ = lambda key,value: None # not allowed

        self.set_property_allowed = True


    def __getattr__(self, item):
        if item == 'refer':
            item = 'referrer'

        if self.keep_url_safe:
            if item in self.safe_fields:
                return getattr(_request, item)
            else:
                return None
        if item in self.url_fields:
            # url 相关的调用, 要先过滤掉prefix
            value = getattr(_request, item, None)
            return get_url_without_prefix(value)
        elif item.startswith('_') and item.lstrip('_') in self.url_fields:
            # 比如 request._path, request._url 返回原始的 _request 上属性
            real_item = item.lstrip('_')
            value = getattr(_request, real_item, None)
            return value
        elif item in self.fields:
            return getattr(_request, item, None)
        elif item == 'url_without_host':
            return _request.url.replace(_request.url_root, '/')
        elif item == 'user_agent':
            user_agent = copy.copy(_request.user_agent)
            user_agent._parser = None
            return user_agent
        elif re.match('^_?path_?\d+$', item): # request.path1
            offset_c = re.search('\d+', item)
            i = offset_c.group()
            return self.get_n_path(i, raw=item.startswith('_'))
        elif re.match('^_?offset_path_?\d+$', item): # request.offset_path_1
            offset_c = re.search('\d+', item)
            i = offset_c.group()
            return self.get_offset_path(i, raw=item.startswith('_'))

        else:
            return self.__dict__.get(item)


    def get_offset_path(self, offset=None, raw=True):
        offset = to_int(offset) or 1
        path = get_path()
        if not raw:
            path = get_url_without_prefix(path)
        return get_request_offset_path(offset, path=path)


    def get_n_path(self, n=0, raw=True):
        # 将 _request.path 按照 / 分割，获得某posiiton上的一部分，不包括 /
        path_parts = _request.path.strip("/").split("/")
        n = to_int(n) -1
        try:
            value = path_parts[n]
        except:
            value = ''
        if raw:
            return value
        else:
            return get_url_without_prefix(value)


    @cached_property
    def raw_user_agent(self):
        return _request.environ.get('HTTP_USER_AGENT') or ''

    @cached_property
    def url_path(self):
        return '/'+get_path()

    @cached_property
    def ip(self):
        return get_visitor_ip()

    @cached_property
    def lang(self):
        return get_language(_request.environ)

    @cached_property
    def language(self):
        return self.lang

    @cached_property
    def protocol(self):
        protocol = _request.environ.get('HTTP_X_PROTOCOL') or 'http'
        return protocol.lower()

    @cached_property
    def is_https(self):
        return self.protocol=='https'

    @cached_property
    def ext(self):
        return self.get_ext()

    @cached_property
    def mime_type(self):
        return self.get_mime_type()


    @cached_property
    def user_agent(self):
        return get_user_agent()


    @staticmethod
    def redirect(url, code=302):
        url = smart_unicode(url)
        force_redirect(url, code=code)


    def get_mime_type(self, path=''):
        if not isinstance(path, (str, unicode)):
            path = smart_unicode(path)
        path = path or get_path()
        if path and isinstance(path, (str, unicode)):
            return guess_type(path) or ''
        else:
            return ''


    def get_ext(self, path=''):
        if not isinstance(path, (str, unicode)):
            path = smart_unicode(path)
        path = path or _request.path
        ext = os.path.splitext(path)[-1] or ''
        ext = ext.lstrip('.').lower()
        return ext

    @staticmethod
    def join(base_url, **kwargs):
        if isinstance(base_url, (str, unicode)):
            return join_url(base_url, **kwargs)
        else:
            return base_url

    @staticmethod
    def decode_url_arg(arg):
        return _decode_url_arg(arg)

    @staticmethod
    def encode_url_arg(arg):
        return _encode_url_arg(arg)


    @cached_property
    def context_doc(self):
        # 上下文对象
        doc_in_g = getattr(g, 'doc', None)
        return doc_in_g



@cache_result
def request():
    return Request()





#coding: utf8
from __future__ import absolute_import
import os
import uuid
from farbox_lite.compiler.posts.posts import get_compiled_md_docs
from farbox_lite.utils import same_slash, smart_unicode, string_types

from .utils import cached, get_sub_docs_by_path, get_sub_docs_by_field, sort_docs_by_field

post_docs_cache = {}

cache_uuid = uuid.uuid4().hex
def get_cache_uuid():
    return cache_uuid

cache_wrapper = cached(cache_key_func=get_cache_uuid)


def get_post_docs(root):
    global cache_uuid
    root = same_slash(root)
    if not os.path.isdir(root):
        return []
    if root in post_docs_cache:
        posts = post_docs_cache.get(root)
        return posts
    posts = get_compiled_md_docs(root) or []
    post_docs_cache[root] = posts
    cache_uuid = uuid.uuid4().hex
    return posts


def update_post_docs(root):
    global cache_uuid
    root = same_slash(root)
    if not os.path.isdir(root):
        return []
    posts = get_compiled_md_docs(root) or []
    post_docs_cache[root] = posts
    cache_uuid = uuid.uuid4().hex
    return posts



@cache_wrapper
def get_posts_by_tag(root, tag, sort=None):
    if not tag:
        return []
    tag = smart_unicode(tag).strip()
    if not tag:
        return []
    raw_posts = get_post_docs(root)
    posts = []
    for post in raw_posts:
        tags = post.get('tags')
        if tags and isinstance(tags, (list, tuple)):
            if tag in tags:
                posts.append(post)
    posts = sort_docs_by_field(posts, field=sort)
    return posts



@cache_wrapper
def get_posts_by_keywords(root, keywords, sort=None):
    if isinstance(keywords, string_types):
        keywords = [keywords]
    keywords = [smart_unicode(keyword) for keyword in keywords]
    keywords = [keyword for keyword in keywords if keyword.strip()]
    raw_posts = get_post_docs(root)
    if not keywords:
        return raw_posts
    posts = []
    for post in raw_posts:
        raw_content = post.get('raw_content')
        if not raw_content:
            continue
        hit = True
        for keyword in keywords:
            if not keyword in raw_content:
                hit = False
                break
        if hit:
            posts.append(post)
    posts = sort_docs_by_field(posts, field=sort)
    return posts


@cache_wrapper
def get_posts_by_field(root, field, value, sort=None):
    raw_posts = get_post_docs(root)
    posts = get_sub_docs_by_field(raw_posts, field, value)
    posts = sort_docs_by_field(posts, field=sort)
    return posts



@cache_wrapper
def get_posts_by_path(root, parent_path, is_direct=False, sort=None):
    raw_posts = get_post_docs(root)
    posts = get_sub_docs_by_path(root, raw_posts, parent_path, is_direct=is_direct)
    posts = sort_docs_by_field(posts, field=sort)
    return posts





#coding: utf8
from __future__ import absolute_import
import sys
import datetime
import traceback
from farbox_lite import version
from .utils import smart_str



def capture_python_errors_to_local_error_log():
    sys.excepthook = when_error_happened


def when_error_happened(e_type, value, tb):
    # config -> settings -> when_error_happened -> configs; will raise error
    try:
        # 错误信息保存到本地
        log_header = '%s @version capture_error %s' % (str(datetime.datetime.now()), version)
        error_filepath = '/tmp/farbox_lite_error_%s.log' % version
        try:
            with open(error_filepath, 'a') as f:
                content_to_write = '\n%s\n' % log_header
                content_to_write = smart_str(content_to_write)
                f.write(content_to_write)
                traceback.print_exception(e_type, value, tb, file=f)
        except:
            pass
        # print into console
        traceback.print_exception(e_type, value, tb)
    except:
        pass


def capture_error(): # 本身不要再引发错误
    try:
        error_info = sys.exc_info()
        if error_info:
            e_type, value, tb = error_info[:3]
            when_error_happened(e_type, value, tb)
    except:
        pass


#coding: utf8
from __future__ import absolute_import

from farbox_lite.utils import smart_unicode
from farbox_lite.utils.lazy import get_value_from_data
from flask import g

social_prefixes = {
    'twitter': 'https://twitter.com',
    'instagram': 'http://instagram.com',
    'dribbble': 'https://dribbble.com',
    'github': 'https://github.com',
    'quora': 'https://quora.com',
    'flickr': 'https://www.flickr.com/photos',
    'weibo': 'http://www.weibo.com',
    'facebook': 'https://www.facebook.com',
    'pinterest': 'http://pinterest.com',
    'vimeo': 'https://vimeo.com',
    'telegram': 'https://telegram.me',
    'youtube': 'https://www.youtube.com',
    'zhihu': 'https://www.zhihu.com/people',
    '500px': 'https://500px.com',
}


social_icons= {
    'twitter': 'fa fa-twitter',
    'instagram': 'fa fa-instagram',
    'dribbble': 'fa fa-dribbble',
    'github': 'fa fa-github',
    'quora': 'fa fa-question-circle-o',
    'flickr': 'fa fa-flickr',
    'weibo': 'fa fa-weibo',
    'facebook': 'fa fa-facebook',
    'pinterest': 'fa fa-pinterest',
    'vimeo': 'fa fa-vimeo',
    'telegram': 'fa fa-send-o',
    'youtube': 'fa fa-youtube',
    '500px': 'fa fa-500px',
}


social_names = social_prefixes.keys()



def get_social_items():
    items = []
    site_configs = get_value_from_data(g, 'site.configs') or {}
    for social_name in social_names:
        username = smart_unicode((site_configs.get(social_name) or '')).strip()
        if not username:
            continue
        social_url_prefix = social_prefixes.get(social_name.lower())
        if '://' in username:
            url = username # 本身就是一个完整的链接
        elif social_url_prefix:
            url = social_url_prefix + '/' + username.split('/')[-1]
        else:
            url = '#'
        item = dict(
            service = social_name,
            username = username,
            url = url,
            icon_class = social_icons.get(social_name) or 'fa fa-link',
        )
        items.append(item)
    return items


def get_social_links():
    social_items = get_social_items()
    if not social_items:
        return ''

    html = ''
    for social_item in social_items:
        html += '<a href="%s" title="%s"><i class="%s"></i></a>\n' % (social_item['url'], social_item['service'], social_item['icon_class'])
    html = '<div class="my_socials">\n%s\n</div>' % html
    return html



#coding: utf8
from __future__ import absolute_import
import re
from flask import g

from farbox_lite.utils import LazyDict, smart_unicode, string_types
from farbox_lite.utils.lazy import to_int, get_value_from_data
from farbox_lite.utils.web.pagination import Paginator
from farbox_lite.utils.web.cache_for_function import cache_result

from .site import get_site_from_request
from .all_docs import has_doc as _has_doc, get_next_doc, get_all_docs_by_field, get_all_docs_by_types, get_doc_by_path, get_all_sub_docs_by_path
from .posts import get_posts_by_field, get_posts_by_tag, get_post_docs, get_posts_by_keywords


MAX_PER_PAGE = 1000


def auto_pg(query, limit, page=None, pager_name=None, with_page=True):
    """
    如果函数是获取列表式的，那么第一个函数可以使用g.pg来获得分页信息。
    """
    if isinstance(query, (dict, int, basestring)): # 本身就是具体的数据了，则不做额外的处理了
        return query

    limit = to_int(limit, 10, max_value=MAX_PER_PAGE) #不能超过 1000

    if not with_page and not page:
        page = 1
    else:
        page = page or getattr(g, 'page', 1)

    paginator = Paginator(query, per_page=limit, page=page)

    if with_page: # 启用分页
        if not hasattr(g, 'paginators'):
            g.paginators = []
        if not hasattr(g, 'paginators_dict'):
            g.paginators_dict = {}
        g.paginators.append(paginator)
        if pager_name:
            g.paginators_dict[pager_name] = paginator

    return paginator.list_object


def get_paginator(index_or_name=0):
    paginators  = list(getattr(g, 'paginators', []))
    paginators_dict = getattr(g, 'paginators_dict', {})
    if paginators and isinstance(index_or_name, int):
        try:
            return paginators[index_or_name]
        except:
            return paginators[0]
    elif paginators_dict and isinstance(index_or_name, (str, unicode)):
        # 必然有paginators
        return paginators_dict.get(index_or_name, paginators[0])
    else:
        return LazyDict()



@cache_result
def get_data(types=None, path=None, url_path=None, limit=None, page=None, with_page=True, keywords=None, pager_name=None, min_limit=0, return_one=False, **kwargs):
    # get site first
    site = kwargs.get('site') or get_site_from_request()

    pager_name = kwargs.get('paginator_name') or pager_name # 两个参数名皆可

    if path and not isinstance(path, (str, unicode)): # 这个类型不支持
        path = None

    if url_path: # 如果使用了url_path，则只会返回一个数据对象，非list
        url_path = smart_unicode(url_path).strip('/')
        return_one = True

    if return_one:
        with_page = False
        limit = 1

    if page is None:
        if not with_page:
            page = 1
        else:
            page = getattr(g, 'page', 1)
    page = to_int(page, 1)

    if not url_path and isinstance(url_path, (str, unicode)): # 指定了url_path，但是空的，返回None
        return None

     # 没有站点时候的数据返回
    if not site:
        if return_one:
            return None
        elif 'return_count' in kwargs:
            return 0
        else:
            return []

    # got root first
    root = site['filepath']

    if not limit:
        limit = kwargs.pop('per_page', None)


    path = path or ''

    if path:
        path = path.lower().lstrip('/')

    # 处理 types，转为 list 类型
    types = types or kwargs.get('type') or 'post'
    if isinstance(types, (str, unicode)):
        types = [doc_type for doc_type in re.split(r' |,|\+', types)]
    elif isinstance(types, (list, tuple)):
        types = [doc_type for doc_type in types if isinstance(doc_type, (str, unicode))]
    else:
        return []

    if len(types)==1 and types[0]=='post':
        is_post_type = True
    else:
        is_post_type = False


    if limit:
        limit = to_int(limit) or 12

    if limit and limit > MAX_PER_PAGE:
        limit = MAX_PER_PAGE


    if not limit: # 设定一个默认输出值
        tmp_per_page = getattr(g, 'tmp_per_page', None)
        if tmp_per_page and isinstance(tmp_per_page, int):
            limit = tmp_per_page
        elif is_post_type:
            limit =  to_int(get_value_from_data(site, 'configs.posts_per_page'), 3)
        else:
            limit = 12

    if min_limit and limit < min_limit:
        limit = min_limit

    if url_path:
        # 根据 url_path 直接返回一个 doc 对象
        matched_posts = get_posts_by_field(root=root, field='url_path', value=url_path)
        if matched_posts:
            return matched_posts[0]
        else:
            return None

    sort = kwargs.get('sort') or None
    tag = kwargs.get('tag')
    tags = kwargs.get('tags')
    if not tag and tags:
        if isinstance(tags, string_types):
            tag = tags
        elif isinstance(tags, (list, tuple)):
            tag = tags[0]
    if keywords:
        docs = get_posts_by_keywords(root=root, keywords=keywords, sort=sort)
    elif tag and isinstance(tag, string_types): # for post only
        docs = get_posts_by_tag(root=root, tag=tag, sort=sort)
    elif len(types) > 1:
        docs = get_all_docs_by_types(root=root, types=types, sort=sort)
    elif len(types) == 1 and types[0]!='all':
        doc_type = types[0]
        docs = get_all_docs_by_field(root=root, field='type', value=doc_type, sort=sort,)
    else:
        docs = get_post_docs(root=root)

    if path:
        is_directly_under = kwargs.get('is_directly_under', False)
        docs = get_all_sub_docs_by_path(root=root, raw_docs=docs, parent_path=path, is_direct=is_directly_under)

    if return_one: # 不返回 list
        if docs:
            return docs[0]
        else:
            return

    if not docs:
        data = []
    else:
        data = auto_pg(docs, limit, page, pager_name, with_page=with_page)

    return data



def get_doc(path=None, match=False, site=None, **kwargs):
    if not site:
        site = get_site_from_request()
    if not site:
        return None
    url_path = kwargs.get('url')
    if url_path:
        doc = get_data(url_path=url_path, site=site)
        return doc
    path = smart_unicode(path)
    root = site['filepath']
    doc = get_doc_by_path(path=path, root=root)
    doc_type_by_user = kwargs.get('type')
    if doc and doc_type_by_user and doc_type_by_user != doc.get('type'):
        # 用户指定了类型，但是类型不匹配，返回 None
        return

    return doc


@cache_result
def has_doc(path, site=None):
    # 一个 doc 是否存在
    if not site:
        site = get_site_from_request()
    if not site:
        return False
    root = site['filepath']
    return _has_doc(root=root, path=path)



@cache_result
def get_connected_one(doc=None, same_folder=None, position='>'):
    doc = doc or getattr(g, 'doc', None)
    if not doc:
        return None
    site = get_site_from_request()
    if not site:
        return None

    root = site['filepath']
    if 'date' not in doc:
        return None
    if same_folder is None and doc.get('type') != 'post': # same_folder未指定，并且是post类型的，则全站
        same_folder = True

    if position == '>': # next_one
        return get_next_doc(current_doc=doc, root=root, sort='-date', same_folder=same_folder)
    else: # pre
        return get_next_doc(current_doc=doc, root=root, sort='date', same_folder=same_folder)




#coding: utf8
from __future__ import absolute_import

def apply_middleware(app, middleware):
    # 处理中间件 & before_request & after_request
    app.wsgi_app = middleware(app.wsgi_app)
    before_request = getattr(middleware, 'before_request',None)
    after_request = getattr(middleware, 'after_request',None)
    if before_request:
        app.before_request_funcs.setdefault(None,[]).append(before_request)
    if after_request:
        app.after_request_funcs.setdefault(None, []).append(after_request)



def apply_before_request(app, func):
    app.before_request_funcs.setdefault(None,[]).append(func)

def apply_after_request(app, func):
    app.after_request_funcs.setdefault(None,[]).append(func)


#coding: utf8
from __future__ import absolute_import
from .web_app import app, web_route_view
from farbox_lite.render.apis.site import create_sites, get_sites



# todo create sites first...

# at last
@app.route('/', methods=['POST', 'GET'])
@app.route('/<path:web_path>', methods=['POST', 'GET'])
def web_route(web_path=''):
    return web_route_view()


# coding: utf8
from __future__ import absolute_import
import datetime
from farbox_lite.utils import smart_str, smart_unicode
from farbox_lite.utils.lazy import to_int, to_float, get_value_from_data


class Date(object):
    def __init__(self, core):
        self.core = core

    def __getattr__(self, item):
        return getattr(self.core, item)

    def __getitem__(self, item):
        # 这样就可以允许用户直接 post.date['%Y-%m']
        if isinstance(item, (unicode, str)):
            return self.format(item)
        else:
            return self.__repr__()

    def __call__(self, *args, **kwargs):
        # 这样就可以允许用户直接 post.date('%Y-%m')
        if len(args) and isinstance(args[0], (str, unicode)):
            return self.format(args[0])
        return self.__repr__()


    def __gt__(self, other):
        if hasattr(other, 'core'):
            other = other.core
        return self.core > other

    def __ge__(self, other):
        if hasattr(other, 'core'):
            other = other.core
        return self.core >= other

    def __lt__(self, other):
        if hasattr(other, 'core'):
            other = other.core
        return self.core < other

    def __le__(self, other):
        if hasattr(other, 'core'):
            other = other.core
        return self.core <= other
    
    def __eq__(self, other):
        if hasattr(other, 'core'):
            other = other.core
        return self.core == other


    def before(self, seconds):
        seconds = to_int(seconds, 0)
        return Date(self.core-datetime.timedelta(seconds=seconds))

    def after(self, seconds):
        seconds = to_int(seconds, 0)
        return Date(self.core+datetime.timedelta(seconds=seconds))

    def format(self, format_str='%H:%M / %d-%m-%Y'):
        if not isinstance(self.core, datetime.datetime):
            return self.core or '-'
        date_offset = 0
        format_str = smart_str(format_str[:200]) # 可以处理中文等unicode字符
        date = self.core + datetime.timedelta(0, date_offset*3600)

        date_year = date.year
        if date_year <= 1900:
            date = date.replace(year=1900)
            result = date.strftime(format_str)
            result = result.replace('1900', str(date_year), 1)
        else:
            result = date.strftime(format_str)
        return smart_unicode(result)

    def __repr__(self):
        return self.format('%Y-%m-%d %H:%M')